import os
import pybedtools
import logging
import math
import pysam
import random
import numpy as np
import array
from GQC import bedtoolslib
from GQC import seqparse

logger = logging.getLogger(__name__)

# Routine choose_bin_size attempts to set the size of bins that will have 
# roughly 1000 primary, non-secondary alignment starts within each bin.
# Input: pysam.AlignmentFile object, pysam.FastaFile object, arguments
# Output: Integer giving the recommended number of bins

def choose_bin_size(alignobj, refobj, includedintervals, args)->int:

    [intervalstart, intervalend] = [0, 0]
    totalreads = 0
    totalrefbases = 0
    for interval in includedintervals:
        ref = interval.chrom
        if ref == "chrM":
            continue
        intervalstart = interval.start
        intervalend = interval.end
        def startsinbin(read):
            if read.reference_start >= intervalstart and read.reference_start < intervalend:
                return True
            else:
                return False

        if args.bincovoverlap:
            readsmapped = alignobj.count(contig = ref, start = intervalstart, end = intervalend)
        else:
            readsmapped = alignobj.count(contig = ref, start = intervalstart, end = intervalend, read_callback = startsinbin)
        totalreads += readsmapped
        totalrefbases += intervalend - intervalstart
        logger.debug("Tallied interval " + ref + ":" + str(intervalstart) + "-" + str(intervalend) + "\t" + str(totalreads) + "\t" + str(totalrefbases))

    arrivalsperbase = totalreads/totalrefbases
    requiredbin = 1000/arrivalsperbase

    # round binsize to nearest 1000 base pairs
    binsize = 1000*math.floor(requiredbin/1000 + 0.5)

    logger.debug("Chose binsize " + str(binsize))

    return binsize
   
def initiate_included_bins(refobj, includedbins, binsize, args)->list:
    bindict = {}
    binnumber = 0

    logger.debug("Initiating bins")
    totallength = 0
    for ref in refobj.references:
        if ref == "chrM":
            continue
        reflength = refobj.get_reference_length(ref)
        totallength += reflength

    numbins = int(totallength/binsize)
    bincounts = array.array('i', (0,)) * numbins
    logger.debug("Zeroed " + str(numbins) + " bins")
    binintervals = []

    #for seqbin in includedbins:
        #if seqbin.end - seqbin.start != binsize:
            #logger.debug("Included interval " + seqbin.chrom + ":" + str(seqbin.start) + "-" + str(seqbin.end) + " is not complete--skipping")
            #continue
        #else:
            #logger.debug("Included interval " + seqbin.chrom + ":" + str(seqbin.start) + "-" + str(seqbin.end) + " is complete--writing count")
            #binname = seqbin.chrom + ":" + str(seqbin.start) 
            #binindex = binindexdict[binname]
            #bincount = coveragebincounts[binindex]
            #cfh.write(seqbin.chrom + "\t" + str(seqbin.start) + "\t" + str(seqbin.end) + "\t" + str(bincount) + "\n")

    for ref in refobj.references:
        logger.debug("Binning " + ref)
        if ref == "chrM":
            continue
        reflength = refobj.get_reference_length(ref)
        binstart = 0
        while binstart + binsize <= reflength:
            binintervals.append(ref + "\t" + str(binstart) + "\t" + str(binstart + binsize) + "\n")
            #binintervalstring = binintervalstring + ref + "\t" + str(binstart) + "\t" + str(binstart + binsize) + "\n"
            bindict[ref + ":" + str(binstart)] = binnumber
            binnumber += 1
            binstart += binsize

    logger.debug("Populated index with " + str(binnumber) + " bin names")

    numbins = len(bincounts)
    bedintervalstring = ''.join(binintervals)
    binintervals = pybedtools.BedTool(binintervalstring, from_string = True)
    numints = len(binintervals)

    logger.debug("Zeroed " + str(numbins) + " bins and created a BedTool object with " + str(numints) + " intervals")

    return [binintervals, bincounts, bindict]

# Bin the genome into binsize-sized intervals (doesn't include mitochondria or ends of chromosomes with fewer than binsize bases)
def initiate_bins(refregionsbed:pybedtools.BedTool, binsize:int, args, includepartial=False)->list:

    bindict = {}
    binnumber = 0

    logger.debug("Initiating bins")
    totallength = 0
    refentries = []
    reflengths = {}
    for region in refregionsbed:
        chrom = region.chrom
        start = region.start
        end = region.end
        if chrom == "chrM":
            continue
        if chrom not in refentries:
            refentries.append(chrom)
            reflengths[chrom] = end
            totallength += end
        else:
            if end > reflengths[chrom]:
                totallength += end - reflengths[chrom]
                reflengths[chrom] = end

    numbins = int(totallength/binsize)
    bincounts = array.array('i', (0,)) * numbins
    logger.debug("Zeroed " + str(numbins) + " bins")
    binintervals = []

    for ref in refentries:
        logger.debug("Binning " + ref)
        if ref == "chrM":
            continue
        reflength = reflengths[ref]
        binstart = 0
        while binstart + binsize <= reflength:
            binintervals.append(ref + "\t" + str(binstart) + "\t" + str(binstart + binsize) + "\n")
            bindict[ref + ":" + str(binstart)] = binnumber
            binnumber += 1
            binstart += binsize
        if includepartial and binstart < reflength:
            binintervals.append(ref + "\t" + str(binstart) + "\t" + str(reflength) + "\n")
            bindict[ref + ":" + str(binstart)] = binnumber

    logger.debug("Populated index with " + str(binnumber) + " bin names")

    numbins = len(bincounts)
    binintervalstring = ''.join(binintervals)
    binintervals = pybedtools.BedTool(binintervalstring, from_string = True)
    numints = len(binintervals)

    logger.debug("Zeroed " + str(numbins) + " bins and created a BedTool object with " + str(numints) + " intervals")

    return [binintervals, bincounts, bindict]

def write_nonexcluded_bin_counts(outputbed:str, includedintervals:pybedtools.BedTool, coveragebins:pybedtools.BedTool, coveragebincounts:list, binindexdict:dict, binsize:int):
    # this should truncate intervals that aren't fully included in the includedintervals:
    includedbins = bedtoolslib.intersectintervals(coveragebins, includedintervals)
    with open(outputbed, "w") as cfh:
        for seqbin in includedbins:
            if seqbin.end - seqbin.start != binsize:
                logger.debug("Included interval " + seqbin.chrom + ":" + str(seqbin.start) + "-" + str(seqbin.end) + " is not complete--skipping")
                continue
            else:
                logger.debug("Included interval " + seqbin.chrom + ":" + str(seqbin.start) + "-" + str(seqbin.end) + " is complete--writing count")
                binname = seqbin.chrom + ":" + str(seqbin.start) 
                binindex = binindexdict[binname]
                bincount = coveragebincounts[binindex]
                cfh.write(seqbin.chrom + "\t" + str(seqbin.start) + "\t" + str(seqbin.end) + "\t" + str(bincount) + "\n")

    return 0

def bin_gccontent_extreme_kmers(refobj:pysam.FastaFile, chrom:str, start:int, end:int, kmersize:int):
    logger.debug("Calculating extreme kmer counts for " + chrom + ":" + str(start) + "-" + str(end))
    refsequence = refobj.fetch(reference=chrom, start=start, end=end).upper()
    [count_at, count_ag, count_ac, count_gc] = [0, 0, 0, 0]
    # also track the GC content in each bin
    [count_gcnucs, count_atnucs] = [0, 0]
    seqstart = 0
    seqstop = kmersize
    # count bases in initial kmer:
    basecounts = {'A':0, 'C':0, 'G':0, 'T':0, 'N':0}
    binsize = end - start
    if seqstop < binsize:
        for i in (range(0, seqstop)):
            thisnuc = refsequence[i]
            basecounts[thisnuc] += 1
            if thisnuc == 'G' or thisnuc == 'C':
                count_gcnucs += 1
            else:
                count_atnucs += 1
        kmerseq = refsequence[0:seqstop]

    while seqstop < binsize:
        if basecounts['N']==0:
            if basecounts['G']==0 and basecounts['C']==0:
                count_at += 1
            if basecounts['A']==0 and basecounts['T']==0:
                count_gc += 1
            if (basecounts['T']==0 and basecounts['C']==0) or (basecounts['A']==0 and basecounts['G']==0):
                count_ag += 1
            if (basecounts['T']==0 and basecounts['G']==0) or (basecounts['A']==0 and basecounts['C']==0):
                count_ac += 1

        # subtract the first base and add the next:
        nextbase = refsequence[seqstop]
        basecounts[kmerseq[0]] -= 1
        basecounts[nextbase] += 1
        kmerseq = kmerseq[1:] + nextbase
        if nextbase == 'G' or nextbase == 'C':
            count_gcnucs += 1
        else:
            count_atnucs += 1

        seqstop += 1

    return [count_at, count_ag, count_ac, count_gc, count_gcnucs, count_atnucs]


def count_extreme_kmers_in_bins(outputbed:str, refobj:pysam.FastaFile, includedintervals:pybedtools.BedTool, coveragebins:pybedtools.BedTool, binsize:int):
    # this should truncate intervals that aren't fully included in the includedintervals:
    includedbins = bedtoolslib.intersectintervals(coveragebins, includedintervals)
    kmersize = 40

    with open(outputbed, "w") as cfh:
        for seqbin in includedbins:
            if seqbin.end - seqbin.start != binsize:
                logger.debug("Bin containing " + seqbin.chrom + ":" + str(seqbin.start) + "-" + str(seqbin.end) + " is partially excluded--skipping")
                continue
            else:
                [count_at, count_ag, count_ac, count_gc, count_gcnucs, count_atnucs] =  bin_gccontent_extreme_kmers(refobj, seqbin.chrom, seqbin.start, seqbin.end, kmersize)
                cfh.write(seqbin.chrom + "\t" + str(seqbin.start) + "\t" + str(seqbin.end) + "\t" + str(count_at) + "\t" + str(count_ag) + "\t" + str(count_ac) + "\t" + str(count_gc) + "\t" + str(count_gcnucs) + "\t" + str(count_atnucs) + "\n")
                logger.debug(seqbin.chrom + "\t" + str(seqbin.start) + "\t" + str(seqbin.end) + "\t" + str(count_at) + "\t" + str(count_ag) + "\t" + str(count_ac) + "\t" + str(count_gc) + "\t" + str(count_gcnucs) + "\t" + str(count_atnucs) + "\n")

    return 0

def tally_chrom_starts_ends(alignobj:pysam.AlignmentFile, refobj:pysam.FastaFile, chrom:str, outputcountbed:str, args):

    chromlength = refobj.get_reference_length(chrom)
    startendcountlist = np.zeros(chromlength, dtype=int)

    for align in alignobj.fetch(chrom):

        refstart = align.reference_start
        refend = align.reference_end

        if align.is_secondary or refstart is None or refend is None:
            continue

        startendcountlist[refstart] += 1
        startendcountlist[refend - 1] -= 1

    print("Done with " + chrom)
    logger.debug("Done with " + chrom)

    return(startendcountlist)

def tally_bin_coverages(alignobj:pysam.AlignmentFile, refobj:pysam.FastaFile, includedintervals:pybedtools.BedTool, includedbed:str, outputcountbed:str, outputincludedcountbed, args):

    if args.covbinsize == 0:
        binsize = choose_bin_size(alignobj, refobj, includedintervals, args)
    else:
        binsize = args.covbinsize

    logger.info("Using bin size " + str(binsize))

    # tally coverage across chromosomes, mosdepth style:
    with open(outputcountbed, "w") as cfh:
        for chrom in alignobj.references:
            startendarray = tally_chrom_starts_ends(alignobj, refobj, chrom, outputcountbed, args)
            # binstart/binend is the first/last zero based position in the current bin
            binstart = 0
            binend = binstart + binsize - 1
            currentreads = 0
            # 0-based position
            currentpos = 0
            chromlength = refobj.get_reference_length(chrom)
            bintotal = 0
    
            while currentpos < chromlength:
                # calculate how many reads are covering this position:
                currentreads = currentreads + startendarray[currentpos]
                bintotal += currentreads
                if currentpos == binend:
                    binaverage = 1.0*bintotal/binsize
                    [count_at, count_ag, count_ac, count_gc, count_gcnucs, count_atnucs] =  bin_gccontent_extreme_kmers(refobj, chrom, binstart, binend + 1, 40)
                    cfh.write(chrom + "\t" + str(binstart) + "\t" + str(binend + 1) + "\t" + str(count_gcnucs) + "\t" + str(binaverage) + "\n")
                    bintotal = 0
                    binstart = binend + 1
                    binend = binstart + binsize - 1
                currentpos += 1

            print("Current reads at end of " + chrom + " is " + str(currentreads))
            logger.debug("Current reads at end of " + chrom + " is " + str(currentreads))

    includedcountfile = bedtoolslib.intersectbed(outputcountbed, includedbed, outputincludedcountbed, requirewhole=True, writefirst=True)
    logger.info("Wrote count bed file of included whole bins")

def tally_included_bin_arrival_rates(alignobj:pysam.AlignmentFile, refobj:pysam.FastaFile, includedintervals:pybedtools.BedTool, outputcountbed:str, args):

    if args.covbinsize == 0:
        binsize = choose_bin_size(alignobj, refobj, includedintervals, args)
    else:
        binsize = args.covbinsize

    # create a BedTools object for the entire reference:
    refbedstring = ""
    for ref in refobj.references:
        reflength = refobj.get_reference_length(ref)
        refbedstring = refbedstring + ref + "\t0\t" + str(reflength) + "\n"
    refbedobj = pybedtools.BedTool(refbedstring, from_string=True)

    logger.info("Using bin size " + str(binsize))
    [coveragebins, coveragebincounts, coveragebinindex] = initiate_bins(refbedobj, binsize, args)

    # this should truncate intervals that aren't fully included in the includedintervals:
    ibin = 0
    kmersize = 40
    includedbins = bedtoolslib.intersectintervals(coveragebins, includedintervals)
    
    [binstart, binend] = [0, 0]
    def startsinbin(read):
        if read.reference_start >= binstart and read.reference_start < binend:
            return True
        else:
            return False

    with open(outputcountbed, "w") as cfh:
        for seqbin in includedbins:
            if seqbin.chrom == "chrM":
                continue
            if seqbin.end - seqbin.start != binsize:
                logger.debug("Included interval " + seqbin.chrom + ":" + str(seqbin.start) + "-" + str(seqbin.end) + " is not complete--skipping")
                continue
            else:
                binchrom = seqbin.chrom
                binstart = seqbin.start
                binend = seqbin.end
                if args.bincovoverlap:
                    bincount = alignobj.count(contig=binchrom, start=binstart, stop=binend)
                else:
                    bincount = alignobj.count(contig=binchrom, start=binstart, stop=binend, read_callback=startsinbin)
                # calculate things about this bin:
                [count_at, count_ag, count_ac, count_gc, count_gcnucs, count_atnucs] =  bin_gccontent_extreme_kmers(refobj, binchrom, binstart, binend, kmersize)
                logger.debug(binchrom + "\t" + str(binstart) + "\t" + str(binend) + "\t" + str(bincount))
                cfh.write(binchrom + "\t" + str(binstart) + "\t" + str(binend) + "\t" + str(count_at) + "\t" + str(count_ag) + "\t" + str(count_ac) + "\t" + str(count_gc) + "\t" + str(count_gcnucs) + "\t" + str(count_atnucs) + "\t" +  str(bincount) + "\n")

def read_extreme_kmer_counts(countfile:str)->dict:

    extremecounts = {}
    with open(countfile, "r") as cfh:
        countline = cfh.readline()
        while countline:
            countline = countline.rstrip()
            [kmertype, count] = countline.split("\t")
            extremecounts[kmertype] = int(count)
            countline = cfh.readline()

    return(extremecounts)

def compare_read_kmers_to_benchmark_kmers(alignobj:pysam.AlignmentFile, refobj:pysam.FastaFile, includedintervals:pybedtools.BedTool, outputbenchmarkcountbed:str, outputreadseqcountbed, outputstrandseqcountbed, config, args):
 
    # Should have this read in from an included count file specified in the config, but for now, calculating it:
    kmersize = args.covkmersize

    if "extremekmercounts" in config.keys() and os.path.exists(config["extremekmercounts"]):
        benchmark_kmer_counts = read_extreme_kmer_counts(config["extremekmercounts"])
    else:
        benchmark_kmer_counts = tally_included_benchmark_kmers(refobj, includedintervals, args)

        with open(outputbenchmarkcountbed, "w") as ofh:
            for kmerseq in benchmark_kmer_counts.keys():
                kmercount = benchmark_kmer_counts[kmerseq]
                ofh.write(kmerseq + "\t" + str(kmercount) + "\n")

    [aligned_read_kmer_counts, stranded_read_kmer_counts] = tally_included_aligned_read_kmers(alignobj, refobj, includedintervals, args)

    with open(outputreadseqcountbed, "w") as ofh:
        for kmerseq in aligned_read_kmer_counts.keys():
            kmercount = aligned_read_kmer_counts[kmerseq]
            if kmerseq in benchmark_kmer_counts.keys():
                benchkmercount = benchmark_kmer_counts[kmerseq]
                kmerratio = 1.0*kmercount/benchkmercount
            else:
                benchkmercount = 'NA'
                kmerratio = 'NA'

            ofh.write(kmerseq + "\t" + str(kmercount) + "\t" + str(benchkmercount) + "\t" + str(kmerratio) + "\n")

    with open(outputstrandseqcountbed, "w") as ofh:
        for kmerseq in stranded_read_kmer_counts.keys():
            kmercount = stranded_read_kmer_counts[kmerseq]
            if kmerseq == 'AT' or kmerseq == 'GC' or kmerseq == 'NA': # own reverse complement, so use total benchmark count
                benchkmercount = benchmark_kmer_counts[kmerseq]
                kmerratio = 1.0*kmercount/benchkmercount
            elif kmerseq in benchmark_kmer_counts.keys():
                benchkmercount = 0.5 * benchmark_kmer_counts[kmerseq]
                kmerratio = 1.0*kmercount/benchkmercount
            else:
                rckmerseq = seqparse.revcomp(kmerseq)
                if rckmerseq in benchmark_kmer_counts.keys():
                    benchkmercount = 0.5 * benchmark_kmer_counts[rckmerseq]
                    kmerratio = 1.0*kmercount/benchkmercount
                else:
                    benchkmercount = 'NA'
                    kmerratio = 'NA'
            ofh.write(kmerseq + "\t" + str(kmercount) + "\t" + str(benchkmercount) + "\t" + str(kmerratio) + "\n")

#
# This function traverses all included intervals (the whole genome or regions specified with --regions, subtracting
# those not in the excluded_regions file), and tallies kmers in the benchmark consensus sequence. chrM is not 
# included due to its unusual copy number.
#
def tally_included_benchmark_kmers(refobj, includedintervals, args)->dict:
    kmercountdict = {}

    k = args.covkmersize
    for interval in includedintervals:
        if interval.chrom == "chrM":
            logger.debug("Skipping mitochondrial region from " + str(interval.start) + " to " + str(interval.end))
            continue
        logger.debug("Calculating kmer counts for " + interval.chrom + ":" + str(interval.start) + "-" + str(interval.end))
        refsequence = refobj.fetch(reference=interval.chrom, start=interval.start, end=interval.end).upper()
        seqstart = 0
        seqstop = k
        # count initial kmer:
        kmerseq = refsequence[0:seqstop]
        # compare to reverse comp and tally
        rckmerseq = seqparse.revcomp(kmerseq)

        intervallength = interval.end - interval.start
        tallykmer = "NA"
        while seqstop <= intervallength:
            if k >= 5: # record extreme kmer types (only 6 different ones since AA==TT, CC==GG, AC==GT and AG==CT for the double stranded benchmark)
                acount = kmerseq.count("A")
                tcount = kmerseq.count("T")
                gcount = kmerseq.count("G")
                ccount = kmerseq.count("C")
                if acount==0 and tcount==0: # kmer is all G's and/or C's
                    if gcount == 0 or ccount == 0: # kmer is entirely G or entirely C
                        tallykmer = "CC"
                    else: # kmer is mix of G's and C's
                        tallykmer = "GC"
                elif gcount==0 and ccount==0: # kmer is all A's and/or T's
                    if acount == 0 or tcount == 0: # kmer is entirely A's or entirely T's
                        tallykmer = "AA"
                    else: # kmer is mix of A's and T's
                        tallykmer = "AT"
                elif (gcount==0 and tcount==0) or (ccount==0 and acount==0): # kmer must be all A's and C's or all G's and T's
                    tallykmer = "AC"
                elif (gcount==0 and acount==0) or (ccount==0 and tcount==0): # kmer must be all A's and G's or all T's and C's
                    tallykmer = "AG"
                else: # kmer has at least three different nucleotides
                    tallykmer = "NA"
            elif kmerseq <= rckmerseq: # if using small kmers (right now, just 3), tally the lower-sorted ("canonical") kmer
                tallykmer = kmerseq
            else:
                tallykmer = rckmerseq
            if tallykmer in kmercountdict.keys():
                kmercountdict[tallykmer] += 1
            else:
                kmercountdict[tallykmer] = 1

            # subtract the first base and add the next:
            if seqstop == intervallength:
                break
            nextbase = refsequence[seqstop]
            kmerseq = kmerseq[1:] + nextbase
            rckmerseq = seqparse.revcomp(kmerseq)
            seqstop += 1

        logger.debug("Last kmer recorded: " + tallykmer)

    return kmercountdict

def tally_included_aligned_read_kmers(alignobj, refobj, includedintervals, args)->dict:
    canonicalkmercountdict = {}
    strandedkmercountdict = {}

    k = args.covkmersize
    minreadalignedpercentage = args.minreadalignedpercentage
    for interval in includedintervals:
        if interval.chrom == "chrM":
            logger.debug("Skipping read alignments in mitochondrial region from " + str(interval.start) + " to " + str(interval.end))
            continue
        logger.debug("Calculating kmer counts for read sequences aligned across " + interval.chrom + ":" + str(interval.start) + "-" + str(interval.end))
        regionstring = interval.chrom + ":" + str(interval.start + 1) + "-" + str(interval.end)

        for align in alignobj.fetch(region=regionstring):
            if align.is_secondary or align.cigartuples is None or (args.downsample is not None and random.random() >= args.downsample):
                continue

            # read sequence is align.query_sequence[query_alignment_start:query_alignment_end], which doesn't include clipped ends
            # it needs to be trimmed to exclude reference positions that are outside of the included region
            alignedreadseq = align.query_alignment_sequence
            querylength = len(alignedreadseq)
            readlength = align.infer_query_length()
            alignrefstart = align.reference_start
            alignrefend = align.reference_end
            if 100*querylength/readlength < minreadalignedpercentage:
                percentclipped = int(100*(readlength - querylength)/readlength)
                logger.debug("Skipping read in " + regionstring + "(" + str(alignrefstart) + "-" + str(alignrefend) + ") due to excessive clipping: " + str(percentclipped) + " percent")
                continue
            cigarops = align.cigartuples
            cigarindex = 0
            while interval.start > alignrefstart:
                logger.debug(str(interval.start) + " is greater than " + str(alignrefstart))
                cigarop = cigarops[cigarindex]
                if cigarop[0] in [0, 1]:
                    alignedreadseq = alignedreadseq[cigarop[1]:]
                if cigarop[0] in [0, 2]:
                    alignrefstart += cigarop[1]
                cigarindex += 1
                if cigarindex >= len(cigarops) - 1:
                    logger.debug("Reached end of cigar ops while trimming alignment start!")
                    break
            cigarindex = len(cigarops) - 1
            while interval.end < alignrefend:
                logger.debug(str(interval.end) + " is less than " + str(alignrefend))
                cigarop = cigarops[cigarindex]
                if cigarop[0] in [0, 1]:
                    alignedreadseq = alignedreadseq[0:-1*cigarop[1]]
                if cigarop[0] in [0, 2]:
                    alignrefend -= cigarop[1]
                cigarindex -= 1
                if cigarindex < 0:
                    logger.debug("Reached beginning of cigar ops while trimming alignment end!")
                    break

            logger.debug("Calculating kmer counts for read " + align.query_name + " from " + align.reference_name + ":" + str(alignrefstart) + "-" + str(alignrefend))
            seqstart = 0
            seqstop = k
            # count initial kmer:
            querylength = len(alignedreadseq)
            kmerseq = alignedreadseq[0:seqstop]
            # compare to reverse comp and tally
            rckmerseq = seqparse.revcomp(kmerseq)
    
            queryreversestrand = align.is_reverse
            tallykmer = "NA"
            while seqstop <= querylength:
                reckmerseq = kmerseq
                recrckmerseq = rckmerseq
                if k >= 5:
                    acount = kmerseq.count("A")
                    tcount = kmerseq.count("T")
                    gcount = kmerseq.count("G")
                    ccount = kmerseq.count("C")
                    if acount==0 and tcount==0: # kmer is all G's and/or C's
                        if gcount==0: # all C's
                            tallykmer = "CC"
                            reckmerseq = "CC"
                            recrckmerseq = "GG"
                        elif ccount == 0: # all G's
                            tallykmer = "CC"
                            reckmerseq = "GG"
                            recrckmerseq = "CC"
                        else: # all G's and C's
                            tallykmer = "GC"
                            reckmerseq = "GC"
                            recrckmerseq = "GC"
                    elif gcount==0 and ccount==0: # kmer is all A's and/or T's
                        if tcount==0: # all A's
                            tallykmer = "AA"
                            reckmerseq = "AA"
                            recrckmerseq = "TT"
                        if acount==0: # all T's
                            tallykmer = "AA"
                            reckmerseq = "TT"
                            recrckmerseq = "AA"
                        else: # all A's and T's
                            tallykmer = "AT"
                            reckmerseq = "AT"
                            recrckmerseq = "AT"
                    elif gcount==0 and tcount==0: # kmer is all A's and C's
                        tallykmer = "AC"
                        reckmerseq = "AC"
                        recrckmerseq = "GT"
                    elif ccount==0 and acount==0: # kmer is all G's and T's
                        tallykmer = "AC"
                        reckmerseq = "GT"
                        recrckmerseq = "AC"
                    elif ccount==0 and tcount==0: # kmer is all A's and G's
                        tallykmer = "AG"
                        reckmerseq = "AG"
                        recrckmerseq = "CT"
                    elif gcount==0 and acount==0: # kmer is all C's and T's
                        tallykmer = "AG"
                        reckmerseq = "CT"
                        recrckmerseq = "AG"
                    else:
                        tallykmer = "NA"
                        reckmerseq = "NA"
                        recrckmerseq = "NA"
                elif kmerseq <= rckmerseq:
                    tallykmer = kmerseq
                else:
                    tallykmer = rckmerseq

                # record the canonical (strand independent) kmer
                if tallykmer in canonicalkmercountdict.keys():
                    canonicalkmercountdict[tallykmer] += 1
                else:
                    canonicalkmercountdict[tallykmer] = 1

                if queryreversestrand:
                    strandedkmer = recrckmerseq
                else:
                    strandedkmer = reckmerseq
                if strandedkmer in strandedkmercountdict.keys():
                    strandedkmercountdict[strandedkmer] += 1
                else:
                    strandedkmercountdict[strandedkmer] = 1
    
                # subtract the first base and add the next:
                if seqstop == querylength:
                    break
                nextbase = alignedreadseq[seqstop]
                kmerseq = kmerseq[1:] + nextbase
                rckmerseq = seqparse.revcomp(kmerseq)
                seqstop += 1

            logger.debug("Last kmer recorded: " + tallykmer + "/" + kmerseq)

    return [canonicalkmercountdict, strandedkmercountdict]




import os
import re
import sys
import random
from tempfile import mkstemp
import pysam
from pysam import VariantFile
import pybedtools
import logging
from GQC import bedtoolslib

logger = logging.getLogger(__name__)

def create_vcf_and_hcbed_from_gvcf(gvcffile:str, args)->int:
    [vcffile, hcbedfile] = ['', '']

    return [vcffile, hcbedfile]

def create_constructed_genome(reffasta:str, vcffile:str, hcbedfile:str, outputdir:str, outputfiles:dict, args)->int:
    samplename = args.samplename

    print("About to construct haplotype specific VCF files")
    [hap1vcf, hap2vcf] = create_haplotype_specific_vcffiles(vcffile, outputdir, outputfiles, args)

    hap1mutfasta = outputfiles['hap1mutfasta']
    hap2mutfasta = outputfiles['hap2mutfasta']
    
    hap1chain = outputfiles['hap1chain']
    hap2chain = outputfiles['hap2chain']

    create_hap_specific_fastafile(reffasta, hap1vcf, samplename + ".hap1", hap1mutfasta, hap1chain, outputdir, outputfiles, args)
    create_hap_specific_fastafile(reffasta, hap2vcf, samplename + ".hap2", hap2mutfasta, hap2chain, outputdir, outputfiles, args)

    hap1hcbed = outputfiles['hap1hcbed']
    hap2hcbed = outputfiles['hap2hcbed']

    hap1unmapped = outputfiles['hap1unmapped']
    hap2unmapped = outputfiles['hap2unmapped']

    hap1lowqualbed = outputfiles['hap1lowqualbed']
    hap2lowqualbed = outputfiles['hap2lowqualbed']

    hap1maskedfasta = outputfiles['hap1maskedfasta']
    hap2maskedfasta = outputfiles['hap2maskedfasta']

    hap1chroms = args.hap1chroms.split(',')
    hap2chroms = args.hap2chroms.split(',')

    lift_and_apply_hq_beds(hcbedfile, hap1chain, hap2chroms, hap1hcbed, hap1unmapped, hap1lowqualbed, hap1mutfasta, hap1maskedfasta, outputdir, outputfiles, args)
    lift_and_apply_hq_beds(hcbedfile, hap2chain, hap1chroms, hap2hcbed, hap2unmapped, hap2lowqualbed, hap2mutfasta, hap2maskedfasta, outputdir, outputfiles, args)

    return 1

def construct_genome_from_reference(refobj, hapvcf:str, hapmutfasta:str, hapchain:str, outputdir:str, outputfiles:dict, args)->list:

    return [hapmutfasta, hapchain]

def create_haplotype_specific_vcffiles(diploidvcf:str, outputdir:str, outputfiles:dict, args)->list:
    samplename = args.samplename

    hap1vcf = outputfiles['hap1vcf']
    hap2vcf = outputfiles['hap2vcf']

    hap1chroms = args.hap1chroms.split(',')
    hap2chroms = args.hap2chroms.split(',')

    vcf_in = VariantFile(diploidvcf)  # auto-detect input format
    vcfheaderstring = str(vcf_in.header)
    matchexp = r"(#CHROM.*" + re.escape(samplename) + ")"
    subsexp1 = r"\1.hap1"
    subsexp2 = r"\1.hap2"
    hap1headerstring = re.sub(matchexp, subsexp1, vcfheaderstring)
    hap2headerstring = re.sub(matchexp, subsexp2, vcfheaderstring)

    hap1headerfd, hap1headerpath = mkstemp()
    with os.fdopen(hap1headerfd, 'w') as fh:
        fh.write(hap1headerstring)
    fh.close()
    hap1vcfheader_in = VariantFile(hap1headerpath)  # auto-detect input format
    os.remove(hap1headerpath)

    hap2headerfd, hap2headerpath = mkstemp()
    with os.fdopen(hap2headerfd, 'w') as fh:
        fh.write(hap2headerstring)
    fh.close()
    hap2vcfheader_in = VariantFile(hap2headerpath)  # auto-detect input format
    os.remove(hap2headerpath)

    vcf_hap1_out = VariantFile(hap1vcf, 'w', header=hap1vcfheader_in.header)
    vcf_hap2_out = VariantFile(hap2vcf, 'w', header=hap2vcfheader_in.header)

    for rec in vcf_in.fetch():
        recsamples = rec.samples.keys()
        hap1chrom = False
        hap2chrom = False
        if rec.contig not in hap2chroms:
            #logger.debug(rec.contig + " is not in hap2 chroms " + args.hap2chroms + " so including in hap 1")
            hap1chrom = True
        if rec.contig not in hap1chroms:
            #logger.debug(rec.contig + " is not in hap1 chroms " + args.hap1chroms + " so including in hap 2")
            hap2chrom = True

        randnum = random.random()
        if hap1chrom:
            allelenum = 1
            newrec = convert_geno_to_haploid(rec, vcf_hap1_out, samplename, samplename + ".hap1", allelenum, randnum)
            vcf_hap1_out.write(newrec)
        if hap2chrom:
            allelenum = 2
            newrec = convert_geno_to_haploid(rec, vcf_hap2_out, samplename, samplename + ".hap2", allelenum, randnum)
            vcf_hap2_out.write(newrec)

    vcf_hap1_out.close()
    vcf_hap2_out.close()

    pysam.tabix_index(hap1vcf, preset="vcf", force=True)
    pysam.tabix_index(hap2vcf, preset="vcf", force=True)

    return [hap1vcf, hap2vcf]

def convert_geno_to_haploid(vcfrecord, vcffile, oldsamplename, newsamplename, allelenum:int, randnum):
    # Create a new record:
    recsamples = vcfrecord.samples.keys()
    newrec = vcffile.new_record(contig=vcfrecord.contig, start=vcfrecord.start, stop=vcfrecord.stop, filter=vcfrecord.filter, alleles = vcfrecord.alleles)

    # is the original genotype here phased or unphased?
    oldsample = vcfrecord.samples[oldsamplename]
    if oldsample.phased:
        if allelenum == 1:
            newallele = oldsample['GT'][0]
        else:
            newallele = oldsample['GT'][-1]
        newrec.samples[newsamplename]['GT'] = (newallele, newallele)
        newrec.samples[newsamplename].phased = True
    else:
        if randnum <= 0.5:
            if allelenum == 1:
                newallele = oldsample['GT'][0]
            else:
                newallele = oldsample['GT'][-1]
        else:
            if allelenum == 1:
                newallele = oldsample['GT'][-1]
            else:
                newallele = oldsample['GT'][-0]
        newrec.samples[newsamplename]['GT'] = (newallele, newallele)
        newrec.samples[newsamplename].phased = False
    
    return newrec

def create_hap_specific_fastafile(reffasta:str, hapvcf:str, hapsamplename:str, hapmutfasta:str, hapchain:str, outputdir:str, outputfiles:dict, args)->int:
    
    bcftoolscmd = "bcftools consensus -c " + hapchain + " -f " + reffasta + " -s " + hapsamplename + " " + hapvcf + " > " + hapmutfasta
    returnvalue = os.system(bcftoolscmd)

    return returnvalue

def lift_and_apply_hq_beds(hcbedfile:str, hcchain:str, hapchroms:list, haphcbed:str, hapunmapped:str, haplcbed:str, hapfasta:str, maskedhapfasta:str, outputdir:str, outputfiles:dict, args)->list:

    # lift the HQ bed to the new haplotype with the bcftools chain file:
    hcname = re.sub(".*/", "", hcbedfile)
    hcout = re.sub(".bed", "", hcname)
    liftcmd = "liftOver " + hcbedfile + " " + hcchain + " " + haphcbed + " " + hapunmapped + " > " + outputdir + "/" + hcout + ".liftover.out"
    logger.debug(liftcmd)
    liftreturnvalue = os.system(liftcmd)

    # subtract lifted bed from lifted genome bed to find low-qual portion:
    hcintervals = pybedtools.BedTool(haphcbed)
    hccorrecthapintervals = []
    for hapinterval in hcintervals:
        if hapinterval.chrom not in hapchroms:
            hccorrecthapintervals.append(hapinterval)

    hapgenomeintervals = bedtoolslib.genomeintervals(hapfasta)
    lcintervals = bedtoolslib.subtractintervals(hapgenomeintervals, hccorrecthapintervals)
    lcintervals.saveas(haplcbed)

    fastamaskcmd = "bedtools maskfasta -fi " + hapfasta + " -bed " + haplcbed + " -fo " + maskedhapfasta
    returnvalue = os.system(fastamaskcmd)

    return returnvalue




import os
import glob
import re
import logging
import importlib.resources

logger = logging.getLogger(__name__)

def plot_benchmark_align_coverage(assemblyname:str, benchname:str, outputdir:str, benchgenomefile:str, benchnlocfile:str, matstring:str):
    rfile_res = importlib.resources.files("GQC").joinpath('BenchCoveragePlot.R')
    with importlib.resources.as_file(rfile_res) as rfile:
        plotcommand = "Rscript " + str(rfile) + " " + assemblyname + " " + benchname + " " + outputdir + " " + benchgenomefile + " " + benchnlocfile + " " + matstring
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_testassembly_align_coverage(assemblyname:str, benchname:str, outputdir:str, resourcedir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('TestCoveragePlot.R')
    with importlib.resources.as_file(rfile_res) as rfile:
        plotcommand = "Rscript " + str(rfile) + " " + assemblyname + " " + outputdir + " " + resourcedir + " " + benchname
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_mononuc_accuracy(assemblyname:str, benchname:str, outputdir:str, resourcedir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('MononucAccuracy.R')
    rlib_res = importlib.resources.files("GQC").joinpath('AssemblyFunctions.R')
    with importlib.resources.as_file(rfile_res) as rfile, importlib.resources.as_file(rlib_res) as rlib:
        plotcommand = "cat " + str(rlib) + " " + str(rfile) + " | " + "Rscript - " + assemblyname + " " + benchname + " " + outputdir + " " + resourcedir
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_qv_score_concordance(assemblyname:str, benchname:str, outputdir:str, resourcedir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('PlotAssemblyQualValueAccuracy.R')
    with importlib.resources.as_file(rfile_res) as rfile:
        plotcommand = "Rscript " + str(rfile) + " " + assemblyname + " " + benchname + " " + outputdir + " " + resourcedir
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_svcluster_align_plots(assemblyname:str, benchname:str, outputdir:str, refobj, mode='bench', prefix=''):

    rfile_res = importlib.resources.files("GQC").joinpath('PlotChromAligns.R')
    if mode == 'compare':
        rfile_res = importlib.resources.files("GQC").joinpath('PlotAssemblyContigAligns.R')

    chromalignbedfiles = glob.glob(outputdir + "/" + prefix + "*.clusters.bed")
    returnvalues = []
    chromdone = {}
    with importlib.resources.as_file(rfile_res) as rfile:
        for chrombed in chromalignbedfiles:
            if mode == 'bench':
                chromosome = chrombed.replace(".clusters.bed", "")
                chromosome = re.sub(r".*/*clustered_aligns\.", "", chromosome)
                chromlength = refobj.get_reference_length(chromosome)
                plotcommand = "Rscript " + str(rfile) + " " + chrombed + " " + assemblyname + " " + benchname + " " + outputdir + " " + str(chromlength)
                logger.debug(plotcommand)
                returnvalues.append(os.system(plotcommand))
            elif mode == 'compare':
                chromosome = chrombed.replace(".clusters.bed", "")
                chromosome = re.sub(r".*/*clustered_aligns\.", "", chromosome)
                if chromosome in chromdone.keys():
                    continue
                chromlength = refobj.get_reference_length(chromosome)
                plotcommand = "Rscript " + str(rfile) + " " + chromosome + " " + assemblyname + " " + benchname + " " + outputdir + " " + str(chromlength)
                logger.debug(plotcommand)
                returnvalues.append(os.system(plotcommand))
                chromdone[chromosome] = True
            else:
                logger.critical("Unknown mode passed to plot_svcluster_align_plots: " + str(mode))
                returnvalues.append(1)

    return returnvalues

def plot_sv_indel_profile_plot(assemblyname:str, benchname:str, outputdir:str, resourcedir:str, refobj):
    rfile_res = importlib.resources.files("GQC").joinpath('IndelProfile.R')

    with importlib.resources.as_file(rfile_res) as rfile:
        plotcommand = "Rscript " + str(rfile) + " " + assemblyname + " " + benchname + " " + outputdir + " " 
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)

    return returnvalue

def plot_assembly_error_stats(assemblyname:str, genomename:str, outputdir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('IndelLengthPlot.R')
    rlib_res = importlib.resources.files("GQC").joinpath('AssemblyFunctions.R')
    with importlib.resources.as_file(rfile_res) as rfile, importlib.resources.as_file(rlib_res) as rlib:
        plotcommand = "cat " + str(rlib) + " " + str(rfile) + " | " + "Rscript - " + assemblyname + " " + genomename + " " + outputdir
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_read_error_stats(readsetname:str, genomename:str, outputdir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('IndelLengthPlot.R')
    rlib_res = importlib.resources.files("GQC").joinpath('AssemblyFunctions.R')
    with importlib.resources.as_file(rfile_res) as rfile, importlib.resources.as_file(rlib_res) as rlib:
        plotcommand = "cat " + str(rlib) + " " + str(rfile) + " | " + "Rscript - " + readsetname + " " + genomename + " " + outputdir
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_read_mononuc_stats(readsetname:str, genomename:str, outputdir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('ReadMononucAccuracy.R')
    rlib_res = importlib.resources.files("GQC").joinpath('ReadBenchPlotFunctions.R')
    with importlib.resources.as_file(rfile_res) as rfile:
        plotcommand = "Rscript " + str(rfile) + " " + readsetname + " " + genomename + " " + outputdir
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_read_coverage_vs_gccontent(readsetname:str, genomename:str, outputdir:str):
    #rfile_res = importlib.resources.files("GQC").joinpath('ReadCoverageVsGCContent.R')
    pass

def plot_read_qv_score_concordance(readsetname:str, benchname:str, outputdir:str, resourcedir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('PlotAssemblyQualValueAccuracy.R')
    with importlib.resources.as_file(rfile_res) as rfile:
        plotcommand = "Rscript " + str(rfile) + " " + readsetname + " " + benchname + " " + outputdir + " " + resourcedir
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def run_ngax_plot(assemblyname:str, benchname:str, outputdir:str, benchgenomebed:str, resourcedir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('NGAxPlot.R')
    rlib_res = importlib.resources.files("GQC").joinpath('AssemblyFunctions.R')
    with importlib.resources.as_file(rfile_res) as rfile, importlib.resources.as_file(rlib_res) as rlib:
        plottitle = "Continuity Curves for " + assemblyname
        plotcommand = "cat " + str(rlib) + " " + str(rfile) + " | " + "Rscript - " + assemblyname + " " + benchname + " " + outputdir + " " + benchgenomebed + " " + plottitle
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_assembly_discrepancy_counts(assemblyname:str, genomename:str, outputdir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('DiscrepancyCountPlot.R')
    rlib_res = importlib.resources.files("GQC").joinpath('AssemblyFunctions.R')
    with importlib.resources.as_file(rfile_res) as rfile, importlib.resources.as_file(rlib_res) as rlib:
        plotcommand = "cat " + str(rlib) + " " + str(rfile) + " | " + "Rscript - " + assemblyname + " " + genomename + " " + outputdir
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_assembly_summary_stats(assemblyname:str, benchname:str, outputdir:str, benchgenomebed:str, resourcedir:str, assemblyqv:int):
    rfile_res = importlib.resources.files("GQC").joinpath('AssemblySummaryPlots.R')
    rlib_res = importlib.resources.files("GQC").joinpath('AssemblyFunctions.R')
    with importlib.resources.as_file(rfile_res) as rfile, importlib.resources.as_file(rlib_res) as rlib:
        plotcommand = "cat " + str(rlib) + " " + str(rfile) + " | " + "Rscript - " + assemblyname + " " + benchname + " " + outputdir + " " + benchgenomebed + " " + str(assemblyqv)
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue

def plot_assembly_comparison_plots(assemblies:list, benchname:str, benchgenomebed:str, outputdir:str):
    rfile_res = importlib.resources.files("GQC").joinpath('AssemblyMultiplot.R')
    rlib_res = importlib.resources.files("GQC").joinpath('AssemblyFunctions.R')
    assemblyfile = outputdir + "/assemblyparams.txt"

    with open(assemblyfile, "w") as afh:
        for assembly in assemblies:
            afh.write(assembly['directory'] + "\t" + assembly['prefix'] + "\t" + assembly['label'] + "\t" + assembly['shortlabel'])
            if 'color' in assembly.keys():
                afh.write("\t" + assembly['color'])
            afh.write("\n")

    with importlib.resources.as_file(rfile_res) as rfile, importlib.resources.as_file(rlib_res) as rlib:
        plotcommand = "cat " + str(rlib) + " " + str(rfile) + " | " + "Rscript - " + assemblyfile + " " + benchname + " " + outputdir + " " + benchgenomebed
        logger.info(plotcommand)
        returnvalue = os.system(plotcommand)
    return returnvalue




import os
import sys
import subprocess
import re
import pysam
import pybedtools
import logging
import math
from pathlib import Path
from collections import namedtuple
from pybedtools import BedTool
from GQC import seqparse
from GQC import kmers
from GQC import bedtoolslib

# create namedtuple for bed intervals:
varianttuple = namedtuple('varianttuple', ['chrom', 'start', 'end', 'name', 'vartype', 'excluded', 'qvscore']) 

logger = logging.getLogger(__name__)

def read_hetsites(hetsitefile)->dict:
    hets = Path(hetsitefile)
    hetsites = {}
    if not hets.is_file():
        logger.error("Het site file " + hetsitefile + " does not exist so phasing analysis will be incomplete")
    else:
        with open(hetsitefile, "r") as hfh:
            hetsiteline = hfh.readline()
            while hetsiteline:
                hetsiteline = hetsiteline.rstrip()
                hetfields = hetsiteline.split("\t")
                chrom = hetfields[0]
                start = hetfields[1]
                end = hetfields[2]
                name = hetfields[3]
                namefields = name.split("_")
                refallele = namefields[-3]
                altallele = namefields[-2]
                hetsitename = chrom + "_" + str(int(start) + 1) + "_" + refallele + "_" + altallele 
                if refallele != "*" and altallele != "*" and len(refallele)==1 and len(altallele)==1:
                    vartype = 'SNV'
                else:
                    vartype = 'INDEL'
                hetsites[hetsitename] = varianttuple(chrom=chrom, start=int(start), end=int(end), name=name, vartype=vartype, excluded=False, qvscore=None)
                hetsiteline = hfh.readline()

    return hetsites

def sort_chrom_hetsite_arrays(hetsites:dict):

    chromhetsites = {}
  
    for hetsite in hetsites.values():
        if hetsite.__class__==dict: #dict
            hetsitetype = 'dict'
            chrom = hetsite['chrom']
            if chrom not in chromhetsites:
                chromhetsites[chrom] = []
            chromhetsites[chrom].append(hetsite)
        else: # tuple
            hetsitetype = 'tuple'
            chrom = hetsite.chrom
            if chrom not in chromhetsites:
                chromhetsites[chrom] = []
            chromhetsites[chrom].append(hetsite)
            
    for chrom in chromhetsites:
        if hetsitetype=="dict":
            chromhetsites[chrom].sort(key=lambda h: (h['start'], h['end']))
        else:
            chromhetsites[chrom].sort(key=lambda h: (h.start, h.end))
    
    return chromhetsites

def write_hetallele_bed(hetsitealleles:dict, hetbed:str):

    contigsortedhetalleles = sort_chrom_hetsite_arrays(hetsitealleles)
    logger.debug("Opening " + hetbed + " to write het alleles along assembly contigs")
    with open(hetbed, "w") as hfh:
        for contig in sorted(contigsortedhetalleles.keys()):
            for hetsite in contigsortedhetalleles[contig]:
                hetname = hetsite['name']
                fields = hetname.split("_")
                strand = fields[-1]
                altallele = fields[-2]
                refallele = fields[-3]
                if hetsite['allele'] == refallele:
                    allelehap = 'SAMEHAP'
                elif hetsite['allele'] == altallele:
                    allelehap = 'ALTHAP'
                else:
                    allelehap = 'OTHER'
                assemblycontig = hetsite['query']
                assemblystart = hetsite['start'] - 1
                assemblyend = hetsite['end'] - 1
                hfh.write(contig + "\t" + str(hetsite['start']) + "\t" + str(hetsite['end']) + "\t" + hetsite['name'] + "\t" + hetsite['allele'] + "\t" + hetsite['ref'] + "\t" + str(hetsite['refstart']) + "\t" + str(hetsite['refend']) + "\t" + assemblycontig + "\t" + str(assemblystart) + "\t" + str(assemblyend) + "\t" + allelehap + "\n")

def map_benchmark_hapmers_onto_assembly(queryfasta, matmarkerfile:str, patmarkerfile:str, outputdir:str, outputfiles:dict):
    #env = os.environ.copy()
    #env['LD_LIBRARY_PATH'] = os.getcwd()
    mathapmeroutput = "KmerMap.mat"
    pathapmeroutput = "KmerMap.pat"
    matpathapmeroutput = "KmerMap.matpat"
    assemblystub = re.sub(r'\.fa.*', "", queryfasta)
    assemblystub = re.sub(r'.*/', "", assemblystub)
    matoutputlocation = outputdir + "/" + mathapmeroutput + "." + assemblystub + ".kmers.merge.bed"
    patoutputlocation = outputdir + "/" + pathapmeroutput + "." + assemblystub + ".kmers.merge.bed"
    matpatoutputlocation = outputdir + "/" + matpathapmeroutput + "." + assemblystub + ".hapmers.merge.bed"
    tmpdir = outputdir + "/tmp"
    if not os.path.exists(outputfiles["phasemarkerbed"]):
        # map maternal/paternal kmers onto the query fasta:
        #matpatoutputlocation = kmers.map_kmer_markers_onto_fasta(queryfasta:str, markerdbs:list, outputdir:str)
        matcommand = "KmerMap -v -m -T2 -P" + tmpdir + " " + matmarkerfile + " " + queryfasta + " " + outputdir + "/" + mathapmeroutput
        patcommand = "KmerMap -v -m -T2 -P" + tmpdir + " " + patmarkerfile + " " + queryfasta + " " + outputdir + "/" + pathapmeroutput
   
        for command in [matcommand, patcommand]:
            path = Path(outputdir + "/tmp")
            #logger.info("Creating temporary directory " + outputdir + "/tmp for output")
            path.mkdir(exist_ok=True)
            #print("Running: " + command)
            logger.info("Running: " + command)
            commandwords = command.split(" ")
            result = subprocess.run(commandwords, capture_output=True, text=True, check=True)
            if result.returncode != 0:
                logger.critical("Command " + command + " had non-zero return code!")
                exit(1)
  
        logger.debug("Merging output files into " + matpatoutputlocation)
        mergecommand = "cat " + matoutputlocation + " " + patoutputlocation + " > " + matpatoutputlocation
        os.system(mergecommand)
    
        logger.debug("Sorting output file " + matpatoutputlocation)
        sortcommand = "sort -k1,1n -k2,2n -k3,3n " + matpatoutputlocation + " > " + outputfiles["phasemarkerbed"]
        os.system(sortcommand)
    else:
        print("Using output file " + outputfiles["phasemarkerbed"])
        logger.info("Using output file " + outputfiles["phasemarkerbed"])

    [mergedhapmerintervals, mergedhapmerbedfile] = bedtoolslib.mergebed(outputfiles["phasemarkerbed"], collapsecolumns='4', collapseoutput='distinct', collapsedelim=',')

    return 0

def map_benchmark_hapmers_onto_assembly_with_phaseblocks(queryfasta, matmarkerfile:str, patmarkerfile:str, outputdir:str, outputfiles:dict):
    env = os.environ.copy()
    env['LD_LIBRARY_PATH'] = os.getcwd()
    phaseblockoutput = "PhaseBlocks"
    assemblystub = re.sub(r'\.fa.*', "", queryfasta)
    assemblystub = re.sub(r'.*/', "", assemblystub)
    outputlocation = outputdir + "/" + phaseblockoutput + "." + assemblystub + ".hapmers.bed"
    tmpdir = outputdir + "/tmp"
    command = "PhaseBlocks -v -T2 -P" + tmpdir + " " + matmarkerfile + " " + patmarkerfile + " " + queryfasta + " " + outputdir + "/" + phaseblockoutput
    if not os.path.exists(outputlocation) and not os.path.exists(outputfiles["phasemarkerbed"]):
        #commandargs = command.split(" ")
        path = Path(outputdir + "/tmp")
        logger.info("Creating temporary directory " + outputdir + "/tmp for output")
        path.mkdir(exist_ok=True)
        print("Running: " + command)
        logger.info("Running: " + command)
        proc = subprocess.Popen(command, shell=True, env=env)
        proc.wait()

    if not os.path.exists(outputfiles["phasemarkerbed"]):
        print("Using output file " + outputlocation)
        logger.info("Using output file " + outputlocation)
        os.rename(outputlocation, outputfiles["phasemarkerbed"])
    else:
        print("Using output file " + outputfiles["phasemarkerbed"])
        logger.info("Using output file " + outputfiles["phasemarkerbed"])

    [mergedhapmerintervals, mergedhapmerbedfile] = bedtoolslib.mergebed(outputfiles["phasemarkerbed"], collapsecolumns='4', collapseoutput='distinct', collapsedelim=',')

    return 0

# This routine duplicates the logic of "MerToPhaseBlock" from Arang Rhie's merqury package (but ported from Java to python)
# I've changed some variable names (e.g., "noBreak" to "includegaps") to make it easier (for me) to read:

def find_phase_blocks_from_marker_bed(bedfile:str, scaffnames:list, shortnum=100, shortlimit=20000, includegaps=True, matpattern='MAT'):

    [mathap, pathap] = find_haplotype_names_in_hapmer_bedfile(bedfile)

    blockbedstring = ""

    pmathap = re.compile(r".*" + re.escape(matpattern) + ".*", re.IGNORECASE)
    matcolor = '255,0,0'
    patcolor = '0,0,255'
    gapcolor = '255,255,255'

    # find scaffold, endpoints, haplotype, and numbers of switches and markers for each phase block
    # also catalog switches (including "Same") in an array to return
    phaseblocks = []
    switches = []

    nummarkers = 0
    numswitches = 0
    numshortswitches = 0
    blockscaff = ""
    blockmarker = ""
    blockstart = -1
    blockend = -1
    shortstart = -1
    shortend = -1
    isshortswitch = True
    with open(bedfile, "r") as bfh:
        markerline = bfh.readline()
        while markerline:
            markerline = markerline.rstrip()
            [scaffold, start, end, markertype] = markerline.split("\t")
            start = int(start)
            end = int(end)

            if markertype == "gap": # gap
                if not includegaps:
                    # create a block with ending at the starting point of the gap, and add it to the list of phaseblocks:
                    blockend = start
                    if blockscaff != "":
                        print("Appending block " + blockscaff + ":" + str(blockstart) + "-" + str(blockend) + " type " + blockmarker + " with " + str(numswitches) + " switches and " + str(nummarkers) + " markers")
                        #phaseblocks.append({'scaffold':blockscaff, 'blockstart':blockstart, 'blockend':blockend, 'blockmarker':blockmarker, 'numswitches':numswitches, 'nummarkers':nummarkers })
                        blockbedstring = blockbedstring + blockscaff + "\t" + str(blockstart) + "\t" + str(blockend) + "\t" + blockmarker + "\t0\t+\t" + str(blockstart) + "\t" + str(blockend) + "\t" + gapcolor + "\n"
                    # start the next block at the end of the gap with "unknown" marker type, without switches and zero markers:
                    blockscaff = scaffnames[int(scaffold)]
                    blockstart = end
                    blockend = end
                    blockmarker = "unknown"
                    numswitches = 0
                    nummarkers = 0
                    isshortswitch = False
            elif (blockscaff == "" or blockscaff != scaffnames[int(scaffold)]): # scaffold has changed or this is start of the first scaff (not a gap line)
                # unless this is the start of the first scaffold, calculate number of markers in the block and add the current block to the list
                if blockscaff != "": # we are currently in a block
                    nummarkers = nummarkers - numshortswitches

                    if pmathap.match(blockmarker):
                        bedcolor = matcolor
                    else:
                        bedcolor = patcolor
                    #print("Appending block " + blockscaff + ":" + str(blockstart) + "-" + str(blockend) + " type " + blockmarker + " with " + str(numswitches) + " switches and " + str(nummarkers) + " markers")
                    #phaseblocks.append({'scaffold':blockscaff, 'blockstart':blockstart, 'blockend':blockend, 'blockmarker':blockmarker, 'numswitches':numswitches, 'nummarkers':nummarkers })
                    blockbedstring = blockbedstring + blockscaff + "\t" + str(blockstart) + "\t" + str(blockend) + "\t" + blockmarker + "\t0\t+\t" + str(blockstart) + "\t" + str(blockend) + "\t" + bedcolor + "\n"
                    # short switches at the ends of scaffolds are included as long range switches
                    if numshortswitches > 0:
                        blockstart = shortstart
                        blockend = shortend
                        nummarkers = numshortswitches
                        blockmarker = markertype
                        if pmathap.match(blockmarker):
                            bedcolor = matcolor
                        else:
                            bedcolor = patcolor
                        print("Appending block " + blockscaff + ":" + str(blockstart) + "-" + str(blockend) + " type " + blockmarker + " with " + str(numswitches) + " switches and " + str(nummarkers) + " markers")
                        #phaseblocks.append({'scaffold':blockscaff, 'blockstart':blockstart, 'blockend':blockend, 'blockmarker':blockmarker, 'numswitches':numswitches, 'nummarkers':nummarkers })
                        blockbedstring = blockbedstring + blockscaff + "\t" + str(blockstart) + "\t" + str(blockend) + "\t" + blockmarker + "\t0\t+\t" + str(blockstart) + "\t" + str(blockend) + "\t" + bedcolor + "\n"
                # start the first block on this new scaffold
                blockmarker = markertype
                numswitches = 0
                numshortswitches = 0
                distfromswitch = 0
                nummarkers = 0
                blockscaff = scaffnames[int(scaffold)]
                blockstart = start
                blockend = end
                if (markertype == mathap or markertype == pathap):
                    nummarkers = nummarkers + 1
                # java version writes out a switch with "Same" here
                #print("SWITCH " + scaffold + ":" + str(start) + "-" + str(end) + " type " + markertype + " switch Same " + " previous type " + blockmarker)
                #switches.append({'scaffold':scaffold, 'start':start, 'end':end, 'markertype':markertype, 'switchtype':'Same', 'blockmarker':blockmarker})
                isshortswitch = False
            else: # same scaffold, not a gap
                nummarkers = nummarkers + 1
                # if the marker type matches the block's type (or the last was an unknown at the beginning of a scaffold), extend block end:
                if blockmarker == markertype or blockmarker == "unknown":
                    blockend = end
                    if blockmarker == "unknown":
                        blockmarker = markertype
                    # if we're in a short switch, weve now come back to the block marker type, so reset shortswitch counter, but first add value to number of switches
                    if isshortswitch:
                        numswitches = numswitches + numshortswitches
                        numshortswitches = 0
                        isshortswitch = False
                        # java version writes out a switch with "SwitchBack" here
                        #print("SWITCH " + blockscaff + ":" + str(start) + "-" + str(end) + " type " + markertype + " switch SwitchBack " + " previous type " + blockmarker)
                        #switches.append({'scaffold':blockscaff, 'start':start, 'end':end, 'markertype':markertype, 'switchtype':'SwitchBack', 'blockmarker':blockmarker})
                    #else:
                        # java version writes out a switch with "Same" here
                        #print("SWITCH " + scaffold + ":" + str(start) + "-" + str(end) + " type " + markertype + " switch Same " + " previous type " + blockmarker)
                        #switches.append({'scaffold':blockscaff, 'start':start, 'end':end, 'markertype':markertype, 'switchtype':'Same', 'blockmarker':blockmarker})
                # if the marker is the opposite of the block's type, extend the short switch and increase the counter
                elif blockmarker != markertype:
                    shortend = end
                    numshortswitches = numshortswitches + 1
                    # if we were not in a short switch already, set the start of the switch region to this marker's start:
                    if not isshortswitch:
                        shortstart = start
                        isshortswitch = True
                        # java version writes out a switch with "Short" here
                        #print("SWITCH " + blockscaff + ":" + str(start) + "-" + str(end) + " type " + markertype + " switch Short " + " previous type " + blockmarker)
                        #switches.append({'scaffold':blockscaff, 'start':start, 'end':end, 'markertype':markertype, 'switchtype':'Short', 'blockmarker':blockmarker})
                    # if we were already in a short switch, check to see if it's numerous or long enough to be a long switch:
                    else:
                        distfromswitch = shortend - shortstart
                        # if we're in a long switch, decrease the number of markers by the number of short switches, and add a block
                        if numshortswitches >= shortnum or distfromswitch > shortlimit:
                            nummarkers = nummarkers - numshortswitches
                            print("Appending block " + blockscaff + ":" + str(blockstart) + "-" + str(blockend) + " type " + blockmarker + " with " + str(numswitches) + " switches and " + str(nummarkers) + " markers")
                            #phaseblocks.append({'scaffold':blockscaff, 'blockstart':blockstart, 'blockend':blockend, 'blockmarker':blockmarker, 'numswitches':numswitches, 'nummarkers':nummarkers})
                            if pmathap.match(blockmarker):
                                bedcolor = matcolor
                            else:
                                bedcolor = patcolor
                            blockbedstring = blockbedstring + blockscaff + "\t" + str(blockstart) + "\t" + str(blockend) + "\t" + blockmarker + "\t0\t+\t" + str(blockstart) + "\t" + str(blockend) + "\t" + bedcolor + "\n"
                            # then initiate a new block on the switched haplotype (allowing the formerly short block to extend)
                            blockstart = shortstart
                            blockend = end
                            nummarkers = numshortswitches
                            numswitches = 0
                            numshortswitches = 0
                            blockmarker = markertype
                            isshortswitch = False
                            # java version writes out a switch with "Long" here
                            #print("SWITCH " + blockscaff + ":" + str(start) + "-" + str(end) + " type " + markertype + " switch Long " + " previous type " + blockmarker)
                            #switches.append({'scaffold':blockscaff, 'start':start, 'end':end, 'markertype':markertype, 'switchtype':'Long', 'blockmarker':blockmarker})
            markerline = bfh.readline()

        #print("Appending block " + blockscaff + ":" + str(blockstart) + "-" + str(end) + " type " + blockmarker + " with " + str(numswitches) + " switches and " + str(nummarkers) + " markers")
        if pmathap.match(blockmarker):
            bedcolor = matcolor
        else:
            bedcolor = patcolor
        blockbedstring = blockbedstring + blockscaff + "\t" + str(blockstart) + "\t" + str(end) + "\t" + blockmarker + "\t0\t+\t" + str(blockstart) + "\t" + str(end) + "\t" + bedcolor + "\n"

    phaseblockbed = pybedtools.BedTool(blockbedstring, from_string = True)

    return phaseblockbed

# Use the Viterbi algorithm to determine the most probable chain of phase block haplotype states #
# given a set of observed haplotype-specific kmers from the benchmark along the test assemblies' scaffolds #
# scaffold fasta object should be passed if you want phase block coordinates to extend to the ends of the 
# scaffolds--otherwise they will only extend to the outermost marker positions!
def find_hapmer_phase_blocks_with_hmm(bedfile:str, hmmphaseblockbedfile:str, scafffastaobj:str, alpha:float, beta:float, distmultiplier=0 ):

    emissionprobs = [[1.0-alpha, alpha], [alpha, 1.0-alpha]]

    # determine maternal/paternal haplotype names in marker bed--these could be mat/pat or hap1/hap2
    [mathap, pathap] = find_haplotype_names_in_hapmer_bedfile(bedfile)

    if mathap is None:
        logger.info("Only haplotype found in " + bedfile + " is " + pathap + "--returning entirely one state")
        mathap = "missinghap"
    if pathap is None:
        logger.info("Only haplotype found in " + bedfile + " is " + mathap + "--returning entirely one state")
        pathap = "missinghap"

    if mathap == "missinghap" and pathap == "missinghap":
        logger.critical("No haplotype markers found for phasing chromosomes!")
        exit(1)

    shortmathap = re.sub("_{0,1}not.*", "", mathap)
    shortmathap = re.sub(r'.*\.', "", shortmathap)
    shortpathap = re.sub(r'_{0,1}not.*', "", pathap)
    shortpathap = re.sub(r'.*\.', "", shortpathap)
    logger.debug("Using " + shortmathap + " for first haplotype name and " + shortpathap + " for second")
    phaseblockbedstring = ""
    mathap1color = '255,0,0'
    pathap2color = '0,0,255'

    haplogprobs = []
    # time 0 has no previous state values:
    previousstate = [[None, None]]
    timepoint = 0

    current_scaffold = ""
    scaffname = ""
    current_pos = 0
    intermarkercoords = []
    phasedend1 = "." + mathap + ".bed"
    phasedend2 = "." + pathap + ".bed"
    matphaseblockbed = hmmphaseblockbedfile.replace('.bed', phasedend1)
    patphaseblockbed = hmmphaseblockbedfile.replace('.bed', phasedend2)
    scaffnamemarkerbed = bedfile.replace('.bed', '.scaffnames.bed')
    mfh = open(matphaseblockbed, "w")
    pfh = open(patphaseblockbed, "w")
    sfh = open(scaffnamemarkerbed, "w")
    scaffnames = scafffastaobj.references
    with open(bedfile, "r") as bfh:
        markerline = bfh.readline()
        while markerline:
            markerline = markerline.rstrip()
            [scaffold, start, end, markertype] = markerline.split("\t")
            scaffold = int(scaffold)
            start = int(start)
            end = int(end)
            midpos = int((start + end)/2)
            scaffname = scaffnames[scaffold]

            # if new midpoint is smaller than current position and on same scaffold, skip it
            # (this is usually because of overlapping regions from different haplogroups)
            if current_scaffold == scaffold and midpos <= current_pos:
                markerline = bfh.readline()
                continue

            if markertype == pathap:
                bedcolor = pathap2color
            else:
                bedcolor = mathap1color
            sfh.write(scaffname + "\t" + str(start) + "\t" + str(end) + "\t" + markertype + "\t0\t+\t" + str(start) + "\t" + str(end) + "\t" + bedcolor + "\n")

            if current_scaffold != scaffold:
                # write out max log prob states if there was a previous chromosome:
                if current_scaffold != "":
                    if len(haplogprobs) > 0:
                        oldscaffname = scaffnames[current_scaffold]
                        lastlogprobs = haplogprobs.pop()
                        scafflength = scafffastaobj.get_reference_length(oldscaffname)
                        write_scaffold_phase_blocks(oldscaffname, scafflength, timepoint, lastlogprobs, previousstate, intermarkercoords, mathap1color, pathap2color, shortmathap, shortpathap, mfh, pfh)
                    timepoint = 0
                current_scaffold = scaffold
                current_pos = 0
                intermarkercoords = []
                previousstate = [[None, None]]
                haplogprobs = []
            if markertype == mathap or markertype == pathap:
                if markertype == mathap:
                    hapindex = 0
                else:
                    hapindex = 1
                # for each timepoint, append an entry to the array of state probabilities--it will be overwritten
                haplogprobs.append([None, None])
                if timepoint == 0:
                    haplogprobs[0][0] = math.log(emissionprobs[0][hapindex])
                    haplogprobs[0][1] = math.log(emissionprobs[1][hapindex])
                    #logger.debug("Assigning haploprobs " + str(haplogprobs[timepoint][0]) + " and " + str(haplogprobs[timepoint][1]) + " at timepoint " + str(timepoint))
                else:
                    previousstate.append([0,0])
                    transitionprobs = [[1.0-beta, beta], [beta, 1.0-beta]]
                    for state in range(2):
                        for laststate in range(2):
                            transitionprob = transitionprobs[laststate][state]
                            newlogprob = haplogprobs[timepoint-1][laststate] + math.log(transitionprob) + math.log(emissionprobs[state][hapindex])
                            if haplogprobs[timepoint][state] is None or newlogprob > haplogprobs[timepoint][state]:
                                haplogprobs[timepoint][state] = newlogprob
                                previousstate[timepoint][state] = laststate
                                #print("Logging previous state " + str(laststate) + " at time " + str(timepoint))
                    #logger.debug("Assigning haploprobs " + str(haplogprobs[timepoint][0]) + " and " + str(haplogprobs[timepoint][1]) + " at timepoint " + str(timepoint))
                intermarkercoords.append([current_pos, midpos])
                current_pos = midpos
                timepoint = timepoint + 1
            markerline = bfh.readline()

    scafflength = scafffastaobj.get_reference_length(scaffname)
    write_scaffold_phase_blocks(scaffname, scafflength, timepoint, lastlogprobs, previousstate, intermarkercoords, mathap1color, pathap2color, shortmathap, shortpathap, mfh, pfh)

    mfh.close()
    pfh.close()
    sfh.close()

    [matblockintervals, matblockbedfile] = bedtoolslib.mergebed(matphaseblockbed, collapsecolumns='4,5,6,7,8,9', collapseoutput='distinct,first,first,first,last,first', collapsedelim=',')
    [patblockintervals, patblockbedfile] = bedtoolslib.mergebed(patphaseblockbed, collapsecolumns='4,5,6,7,8,9', collapseoutput='distinct,first,first,first,last,first', collapsedelim=',')

    allphaseblockints = matblockintervals.cat(patblockintervals, postmerge=False).sort()

    return allphaseblockints

def find_haplotype_names_in_hapmer_bedfile(bedfile:str):

    # if one found haplotype matches ".mat" or "[ab]1_not_[ab]2" (case insensitive), return it first--covers both benchmark and assembly comparison
    pmathap = re.compile(r'.*\.mat.*|[qr]1_not_[qr]2', re.IGNORECASE)
    pcomma = re.compile(r'.*\,.*', re.IGNORECASE)
    # determine two haplotypes in the "name" field of the input bed file (assumes only two distinct values plus optionally "gap")
    with open(bedfile, "r") as bfh:
        markerline = bfh.readline()
        hap1 = None
        hap2 = None
        while markerline:
            markerline = markerline.rstrip()
            [scaffold, start, end, markertype] = markerline.split("\t")
            if pcomma.match(markertype):
                logger.debug("Skipping marker name " + markertype)
                markerline = bfh.readline()
                continue
            if markertype != "gap" and hap1 is None:
                hap1 = markertype
                logger.debug("Found first marker name " + markertype)
            elif markertype != "gap" and markertype != hap1:
                hap2 = markertype
                logger.debug("Found second marker name " + markertype)
                break
            markerline = bfh.readline()
    if hap1 is not None and hap2 is not None:
        if pmathap.match(hap1):
            logger.info("Found hap1  " + hap1 + " and hap2 " + hap2 + " as haplotypes.")
            print("Found hap1 " + hap1 + " and hap2 " + hap2 + " as haplotypes.")
            return [hap1, hap2]
        elif pmathap.match(hap2):
            logger.info("Found hap2 " + hap2 + " and hap1  " + hap1 + " as haplotypes.")
            print("Found hap2 " + hap2 + " and  hap1 " + hap1 + " as haplotypes.")
            return [hap2, hap1]
        else:
            logger.info("Unable to determine mat/pat or 1/2 identities of hap1 " + hap1 + " and hap2 " + hap2)
            print("Unable to determine mat/pat or 1/2 identities of hap1 " + hap1 + " and hap2 " + hap2)
            exit(1)
    else:
        logger.info("Unable to find two haplotypes in " + bedfile)
        print("Unable to find two haplotypes in " + bedfile)
        if hap1 is not None:
            if pmathap.match(hap1):
                return[hap1, hap2]
            else:
                return [hap2, hap1]
        if hap2 is not None:
            if pmathap.match(hap2):
                return [hap2, hap1]
            else:
                return [hap1, hap2]

        return [hap1, hap2]

def write_scaffold_phase_blocks(scaffoldname:str, scafflength:int, timepoint:int, lastlogprobs:list, previousstate:list, intermarkercoords:list, hap1color:str, hap2color:str, mathap:str, pathap:str, hap1fh, hap2fh):
    prevstatelength = len(previousstate)
    intermarkerlength = len(intermarkercoords)
    path = [0] * timepoint
    if lastlogprobs[0] > lastlogprobs[1]:
        path[timepoint-1] = 0
    else:
        path[timepoint-1] = 1
    revtimepoints = range(timepoint-2, -1, -1)
    for currenttime in revtimepoints:
        path[currenttime] = previousstate[currenttime + 1][path[currenttime + 1]]
    for currenttime in range(timepoint):
        if path[currenttime] == 0:
            currenthap = mathap
            fh = hap1fh
            bedcolor = hap1color
        else:
            currenthap = pathap
            fh = hap2fh
            bedcolor = hap2color

        if currenttime == 0:
            blockstart = 0
        else:
            blockstart = int((intermarkercoords[currenttime][0] + intermarkercoords[currenttime][1])/2.0)

        if currenttime < timepoint-1:
            blockend = int((intermarkercoords[currenttime+1][0] + intermarkercoords[currenttime+1][1])/2.0)
        else:
            blockend = scafflength
        fh.write(scaffoldname + "\t" + str(blockstart) + "\t" + str(blockend) + "\t" + currenthap + "\t0\t+\t" + str(blockstart) + "\t" + str(blockend) + "\t" + bedcolor + "\n")

    return 0



import re
import os
import pysam
import pybedtools
import logging
import statistics
from collections import namedtuple
from pathlib import Path
from GQC import seqparse
from GQC import phasing
from GQC import bedtoolslib
from GQC import output

# create namedtuple for bed intervals:
varianttuple = namedtuple('varianttuple', ['chrom', 'start', 'end', 'name', 'vartype', 'excluded', 'qvscore']) 

logger = logging.getLogger(__name__)

# query start is always less than query end regardless of strand. query left always corresponds to ref start, 
# and so will be greater than query right when strand is reversed--all four are 1-based
def write_bedfiles(bamobj, pafaligns, refobj, queryobj, hetsites, testmatbed, testpatbed, truthbed, hetallelebed, excludedbedobj, benchparams, args):

    refcoveredstring = ""
    querycoveredstring = ""
    refcoveredbed = None
    querycoveredbed = None
    if 'variantfile' in args:
        user_variantfile = args.variantfile
    else:
        user_variantfile = None
    variants = []
    hetsitealleles = {}
    alignedscorecounts = []
    snverrorscorecounts = []
    indelerrorscorecounts = []

    if bamobj is not None:
        for align in bamobj.fetch():
            if align.is_secondary:
                continue
            if align.reference_length >= args.minalignlength:
                query, querystart, queryend, ref, refstart, refend, strand = retrieve_align_data(align)
                if strand == "F":
                    queryleft = querystart
                    queryright = queryend
                else:
                    queryleft = queryend
                    queryright = querystart
                querynamestring = query + "." + str(queryleft) + "." + str(queryright)
                refnamestring = ref + "." + str(refstart) + "." + str(refend) + "." + strand
                querycoveredstring += query + "\t" + str(querystart - 1) + "\t" + str(queryend) + "\t" + refnamestring + "\n"
                refcoveredstring += ref + "\t" + str(refstart - 1) + "\t" + str(refend) + "\t" + querynamestring + "\n"
                if user_variantfile is None:
                    variants.extend(align_variants(align, queryobj, query, querystart, queryend, refobj, ref, refstart, refend, strand, hetsites, hetsitealleles, alignedscorecounts, snverrorscorecounts, indelerrorscorecounts, True))
        # mark variants that are in excluded regions:
        logger.debug("Beginning to exclude variants in excluded regions")
        if excludedbedobj:
            variants = exclude_variants(variants, excludedbedobj)
            logger.debug("Finished excluding variants in excluded regions")
    else:
        for pafdict in pafaligns:
            # all start/endpoints are 1-based
            query = pafdict['query']
            querystart = pafdict['querystart']
            queryend = pafdict['queryend']
            if querystart < queryend:
                queryleft = querystart
                queryright = queryend
            else:
                queryleft = queryend
                queryright = querystart
            ref = pafdict['target']
            refstart = pafdict['targetstart']
            refend = pafdict['targetend']
            strand = pafdict['strand']

            querynamestring = query + "." + str(querystart) + "." + str(queryend)
            refnamestring = ref + "." + str(refstart) + "." + str(refend) + "." + strand
            querycoveredstring += query + "\t" + str(queryleft - 1) + "\t" + str(queryright) + "\t" + refnamestring + "\n"
            refcoveredstring += ref + "\t" + str(refstart - 1) + "\t" + str(refend) + "\t" + querynamestring + "\n"

    refcoveredbed = pybedtools.BedTool(refcoveredstring, from_string = True)
    querycoveredbed = pybedtools.BedTool(querycoveredstring, from_string = True)

    # if testpatbed is None, a single bed file should be written to the file passed as testmatbed:
    if not os.path.exists(testmatbed) or (testpatbed is not None and not os.path.exists(testpatbed)):
        matpattern = benchparams['matpattern']
        patpattern = benchparams['patpattern']
        phap1 = re.compile(r".*" + re.escape(matpattern) + ".*", re.IGNORECASE)
        phap2 = re.compile(r".*" + re.escape(patpattern) + ".*", re.IGNORECASE)
        with open(testmatbed, "w") as tmb:
            for testint in sorted(querycoveredbed, key=lambda h: (h.chrom, h.start, h.stop)):
                if testpatbed is None or phap1.match(testint.name):
                    tmb.write(testint.chrom + "\t" + str(testint.start) + "\t" + str(testint.end) + "\t" + testint.name + "\n")

        if testpatbed is not None:
            with open(testpatbed, "w") as tpb:
                for testint in sorted(querycoveredbed, key=lambda h: (h.chrom, h.start, h.stop)):
                    if phap2.match(testint.name):
                        tpb.write(testint.chrom + "\t" + str(testint.start) + "\t" + str(testint.end) + "\t" + testint.name + "\n")

    if not os.path.exists(truthbed) or os.path.getsize(truthbed) == 0:
        with open(truthbed, "w") as rb:
            for truthint in sorted(refcoveredbed, key=lambda h: (h.chrom, h.start, h.stop)):
                rb.write(truthint.chrom + "\t" + str(truthint.start) + "\t" + str(truthint.end) + "\t" + truthint.name + "\n")

    if hetsites and (hetallelebed is not None) and (not os.path.exists(hetallelebed) or (os.path.getsize(hetallelebed) == 0)):
        phasing.write_hetallele_bed(hetsitealleles, hetallelebed)

    if user_variantfile is not None:
        with open(user_variantfile, "r") as vh:
            variantline = vh.readline()
            while variantline:
                variantline = variantline.rstrip()
                [chrom, start, end, name] = variantline.split("\t")
                variants.append(varianttuple(chrom=chrom, start=int(start), end=int(end), name=name, qvscore=None))
                variantline = vh.readline()

    return [refcoveredbed, querycoveredbed, variants, hetsitealleles, alignedscorecounts, snverrorscorecounts, indelerrorscorecounts]

def retrieve_align_data(align)->list:
    if align.is_reverse:
        strand = 'R'
    else:
        strand = 'F'
    query = align.query_name

    # find number of hard clipped bases:
    cigartuples = align.cigartuples
    # will use actual one-based positions, so I don't go crazy, then report BED format zero-based half open later
    # pysam's align.query_alignment_start is the 0-based coordinate within the uncomplemented hard clipped query sequence, so here we add hardclipping from the 
    # appropriate end to get coordinates within the entire sequence (so low coordinates are towards the start of the original query sequence, not the left end
    # of the alignment)
    if strand == 'F':
        if cigartuples[0][0] == 5:
            hardclip = cigartuples[0][1]
        else:
            hardclip = 0
        querystart = hardclip + align.query_alignment_start + 1 # first unclipped base (was 0-based, now 1-based)
        queryend = hardclip + align.query_alignment_end # was 0-based index just past the last unclipped base now 1-based last unclipped base
    elif strand == 'R':
        if cigartuples[-1][0] == 5:
            hardclip = cigartuples[-1][1]
        else:
            hardclip = 0
        querystart = hardclip + align.query_length - align.query_alignment_end + 1 # was 0-based index just past end of reversed query, now 1-based 5' end of query
        queryend = hardclip + align.query_length - align.query_alignment_start # was 0-based index of first unclipped base, now 1-based 3' unclipped end of query

    ref = align.reference_name

    refstart = align.reference_start + 1 # was zero-based, now one-based
    refend = align.reference_end #  was zero-based, adding one for bed format "half open", now one-based

    aligndata = [query, querystart, queryend, ref, refstart, refend, strand]

    return aligndata

# query start and query end are the lower and higher endpoints of the query seq in query coordinates (1-based)
# regardless of orientation of the alignment
def align_variants(align, queryobj, query:str, querystart:int, queryend:int, refobj, ref:str, refstart:int, refend:int, strand:str, chromhetsites={}, hetsitealleles={}, alignedscorecounts=[], snverrorscorecounts=[], indelerrorscorecounts=[], widen=True)->list:

    # coordinates are all one-based, with start at beginning of *original* sequence (not left end of the alignment)
    variantlist = []
    coveredregionlist = []
    homozygousregionlist = []

    # make an array of query positions for each ref position:
    query_positions = []

    if queryobj is None:
        queryseq = align.query_alignment_sequence
    else:
        queryseq = queryobj.fetch(reference=query, start=querystart-1, end=queryend).upper()

    # qual scores might be None, but if they're not, add to the counts in alignedscorecounts and errorscorecounts:
    alignedqualscores = align.query_alignment_qualities

    if alignedqualscores is not None:
        if len(alignedscorecounts)==0:
            for i in range(100):
                alignedscorecounts.append(0)
                snverrorscorecounts.append(0)
                indelerrorscorecounts.append(0)
        for qualscorechar in alignedqualscores:
            qualscore = int(qualscorechar)
            alignedscorecounts[qualscore] = alignedscorecounts[qualscore] + 1

    refseq = refobj.fetch(reference=ref, start=refstart-1, end=refend).upper()

    bedstrand = '+'
    if strand == 'R':
        if queryobj is not None:
            queryseq = seqparse.revcomp(queryseq)
        bedstrand = '-'

    alignops = align.cigartuples

    # first position will begin at left-most ref/query base of the alignment (regardless of strand)
    refcurrentoffset = 0
    querycurrentoffset = 0

    alignopindex = 0
    refalignlength = refend - refstart + 1
    queryalignlength = queryend - querystart + 1
    matchns = re.compile(".*[nN].*")
    numops = len(alignops)

    #logger.debug("Traversing alignment with " + str(numops) + " cigar ops aligning " + ref + " to " + query)
    while refcurrentoffset <= refend-refstart and alignopindex < len(alignops): # traverse the alignment operator by operator
        alignop = alignops[alignopindex]
        op = alignop[0]
        oplength = alignop[1]

        if op in [0, 7, 8]: # MX= find SNV and MNVs
            for blockoffset in range(oplength):
                refpos = refcurrentoffset + blockoffset # this is distance from left-most base of the alignment
                querypos = querycurrentoffset + blockoffset # this is distance from left-most base of the alignment
                if strand == 'F':
                    querycoordinate = querypos + querystart
                else:
                    querycoordinate = queryend - querypos
                if refpos >= len(refseq):
                    logger.debug("Ref position " + str(refpos) + " in " + ref + " is past end of seq")
                    continue
                if querypos >= len(queryseq):
                    logger.debug("Query position " + str(querypos) + " in " + query + " is past end of seq")
                    continue
                if refseq[refpos] != queryseq[querypos] and refseq[refpos] != "N" and queryseq[querypos] != "N":
                    variantname=query+"_"+str(querycoordinate)+"_"+refseq[refpos]+"_"+queryseq[querypos]+"_"+strand # query's 1-based position, ref base, query base (comp if rev strand), strand
                    #additionalfields = "0\t" + bedstrand + "\t" + str(refpos+refstart-1) + "\t" + str(refpos+refstart) + "\t0,0,0\t" + alignstring
                    if alignedqualscores is not None:
                        snverrorscore = int(alignedqualscores[querypos])
                        snverrorscorecounts[snverrorscore] = snverrorscorecounts[snverrorscore] + 1
                    else:
                        snverrorscore = None
                    variantlist.append(varianttuple(chrom=ref, start=refpos+refstart-1, end=refpos+refstart, name=variantname, vartype='SNV', excluded=False, qvscore=snverrorscore))
                elif queryseq[querypos] == 'N':
                    if alignedqualscores is not None:
                        snverrorscore = int(alignedqualscores[querypos])
                        snverrorscorecounts[snverrorscore] = snverrorscorecounts[snverrorscore] + 1

                query_positions.append(querypos)

        if op in [2, 3]: # deletions
            refpos = refcurrentoffset-1;
            for deloffset in range(oplength):
                query_positions.append(querycurrentoffset)
            refallele = refseq[refcurrentoffset:refcurrentoffset+oplength] # one-based refstart+refcurrentoffset to refstart+refcurrentoffset+oplength-1
            queryallele = "*" # one-based between querystart+querycurrentoffset-1 and querystart+querycurrentoffset if forward strand, queryend-querycurrentoffset+1 and queryend-querycurrentoffset if rev
            extendleft = 0
            extendright = 0
            if widen is True: # n.b. - this will *lower* the righthand coordinate of reverse strand queries by "extendright" and *lower* the lefthand coordinate of ref entries and forward strand queries by "extendleft"
                while querycurrentoffset + extendright < queryalignlength and refcurrentoffset + extendright < refalignlength and refseq[refcurrentoffset + extendright] == queryseq[querycurrentoffset + extendright]:
                    refallele = refallele + queryseq[querycurrentoffset + extendright]
                    if queryallele == "*":
                        queryallele = queryseq[querycurrentoffset + extendright]
                    else:
                        queryallele = queryallele + queryseq[querycurrentoffset + extendright]
                    extendright = extendright + 1
                while querycurrentoffset - 1 - extendleft >= 0 and refcurrentoffset - 1 + oplength - extendleft >= 0 and refseq[refcurrentoffset - 1 + oplength - extendleft] == queryseq[querycurrentoffset - 1 - extendleft]:
                    refallele = queryseq[querycurrentoffset - extendleft - 1] + refallele
                    if queryallele == "*":
                        queryallele = queryseq[querycurrentoffset - extendleft - 1]
                    else:
                        queryallele = queryseq[querycurrentoffset - extendleft - 1] + queryallele
                    extendleft = extendleft + 1
            
            # querycoordinate values still need to be adjusted due to widening?
            if strand == 'F':
                querycoordinate = querystart + querycurrentoffset - extendleft
            else:
                querycoordinate = queryend - querycurrentoffset - extendright

            # if there are quality scores, find the median quality across this deletion for our error tally (and to report in the output)
            indelerrorscore = None
            if alignedqualscores is not None:
                if extendleft == 0 and extendright == 0:
                    queryquals = alignedqualscores[querycurrentoffset-1:querycurrentoffset]
                    qualscores = [int(x) for x in queryquals]
                else:
                    queryquals = alignedqualscores[querycurrentoffset-extendleft:querycurrentoffset+extendright]
                    qualscores = [int(x) for x in queryquals]
                if len(qualscores)==0:
                    variantname = variantname + ""
                    #logger.debug("Variant with pos " + ref + ":" + str(refpos+refstart-extendleft) + "-" + str(refpos+refstart+oplength+extendright) + " name " + variantname + " and queryallele " + queryallele + " and refallele " + refallele + " has query surrounding seq " + querysurroundingseq + " and length of qualscores is zero!")
                else:
                    numquals = len(qualscores)
                    # if even number of qual scores, drop the top one so the lower of the two medians is chosen (rather than an average, which may not be represented in the total qv score counts)
                    if numquals==2*int(numquals/2):
                        qualscores.pop()
                    medqual = int(statistics.median(qualscores))
                    indelerrorscorecounts[medqual] = indelerrorscorecounts[medqual] + 1
                    indelerrorscore = medqual

            # check neighboring bases for Ns, which can create misleading/possibly wrong variants
            # (looks as if I might be looking only at left and right bases within the query alleles here--need to check)
            [queryleftbase, queryrightbase] = ["", ""]
            if querycurrentoffset > 0:
                queryleftbase = queryseq[querycurrentoffset-1]
            if querycurrentoffset+extendright < queryalignlength:
                queryrightbase = queryseq[querycurrentoffset+extendright]
            querysurroundingseq = queryleftbase + queryrightbase
        
            if not (matchns.match(queryallele) or matchns.match(refallele) or matchns.match(querysurroundingseq)):
                variantname=query+"_"+str(querycoordinate)+"_"+refallele+"_"+queryallele+"_"+strand # positions of insertions are positions to the left of first inserted base

                variantlist.append(varianttuple(chrom=ref, start=refpos+refstart-extendleft, end=refpos+refstart+oplength+extendright, name=variantname, vartype='INDEL', excluded=False, qvscore=indelerrorscore ))
            else:
                variantname=query+"_"+str(querycoordinate)+"_"+refallele+"_"+queryallele+"_"+strand # positions of insertions are positions to the left of first inserted base
                #logger.debug("Variant with name " + variantname + " and queryallele " + queryallele + " and refallele " + refallele + " has query surrounding seq " + querysurroundingseq + " was excluded")

        if op == 1: # insertion
            refpos = refcurrentoffset-1;
            refallele = "*"
            queryallele = queryseq[querycurrentoffset:querycurrentoffset+oplength] # one-based querystart+querycurrentoffset to querystart+querycurrentoffset+oplength-1 if forward strand, queryend-querycurrentoffset to queryend-querycurrentoffset-oplength+1 if reverse
            if alignedqualscores is not None:
                queryquals = alignedqualscores[querycurrentoffset:querycurrentoffset+oplength]
                qualscores = [int(x) for x in queryquals]
            #query_positions.append(querycurrentoffset)
            extendright = 0
            extendleft = 0
            if widen is True: # n.b. - this will *lower* the righthand coordinate of reverse strand queries by "extendright"
                while querycurrentoffset + extendright < queryalignlength and refcurrentoffset + extendright < refalignlength and refseq[refcurrentoffset + extendright] == queryseq[querycurrentoffset + extendright]:
                    queryallele = queryallele + refseq[refcurrentoffset + extendright]
                    if alignedqualscores is not None:
                        qualscores.append(int(alignedqualscores[querycurrentoffset + extendright]))
                    if refallele == "*":
                        refallele = refseq[refcurrentoffset + extendright]
                    else:
                        refallele = refallele + refseq[refcurrentoffset + extendright]
                    extendright = extendright + 1
                while querycurrentoffset - 1 + oplength - extendleft >= 0 and refcurrentoffset - 1 - extendleft >= 0 and refseq[refcurrentoffset - 1 - extendleft] == queryseq[querycurrentoffset - 1 + oplength - extendleft]:
                    queryallele = refseq[refcurrentoffset - extendleft - 1] + queryallele
                    if alignedqualscores is not None:
                        qualscores.insert(0, int(alignedqualscores[querycurrentoffset - 1 + oplength - extendleft]))
                    if refallele == "*":
                        refallele = refseq[refcurrentoffset - extendleft - 1]
                    else:
                        refallele = refseq[refcurrentoffset - extendleft - 1] + refallele
                    extendleft = extendleft + 1
            if strand == 'F':
                querycoordinate = querystart + querycurrentoffset - extendleft
                querycoordend = querycoordinate + oplength - 1 # this is potentially off by one and could be a bug (see its use below)
            else:
                querycoordinate = queryend - querycurrentoffset - extendleft
                querycoordend = querycoordinate - oplength - 1 # this is potentially off by one and could be a bug (see its use below)

            # if there are quality scores, find the median quality across this insertion for our error tally (and to report in the output)
            indelerrorscore = None
            if alignedqualscores is not None:
                numquals = len(qualscores)
                # if even number of qual scores, drop the top one so the lower of the two medians is chosen (rather than an average, which may not be represented in the total qv score counts)
                if numquals==2*int(numquals/2):
                    qualscores.pop()
                medqual = int(statistics.median(qualscores))
                indelerrorscorecounts[medqual] = indelerrorscorecounts[medqual] + 1
                indelerrorscore = medqual

            # check neighboring bases for Ns, which can create misleading/possibly wrong variants
            [refleftbase, refrightbase] = ["", ""]
            if refcurrentoffset-extendleft > 0:
                refleftbase = refseq[refcurrentoffset-1-extendleft]
            else:
                refleftbase = ''
            if refcurrentoffset+extendright+1 < refalignlength:
                refrightbase = refseq[refcurrentoffset+extendright+1]
            else:
                refrightbase = ''
            refsurroundingseq = refleftbase + refrightbase
            
            if not (matchns.match(queryallele) or matchns.match(refallele) or matchns.match(refsurroundingseq)):
                variantname=query+"_"+str(querycoordinate)+"_"+refallele+"_"+queryallele+"_"+strand
                #additionalfields = "0\t" + bedstrand + "\t" + str(refpos+refstart) + "\t" + str(refpos+refstart+extendright) + "\t0,0,0\t" + alignstring
                #logger.debug("Variant with pos " + ref + ":" + str(refpos+refstart-extendleft) + "-" + str(refpos+refstart+extendright) + " name " + variantname + " and queryallele " + queryallele + " and refallele " + refallele + " has ref surrounding seq " + refsurroundingseq)
                variantlist.append(varianttuple(chrom=ref, start=refpos+refstart-extendleft, end=refpos+refstart+extendright, name=variantname, vartype='INDEL', excluded=False, qvscore=indelerrorscore ))
            else:
                variantname=query+"_"+str(querycoordinate)+"_"+refallele+"_"+queryallele+"_"+strand
                #logger.debug("Variant with name " + variantname + " and queryallele " + queryallele + " and refallele " + refallele + " has ref surrounding seq " + refsurroundingseq + " was excluded")

        # advance current positions: cases where reference coord advances (MDN=X):
        if op in [0, 2, 3, 7, 8]:
            refcurrentoffset = refcurrentoffset + oplength
        # cases where query coord advances (MI=X)
        if op in [0, 1, 7, 8]:
            querycurrentoffset = querycurrentoffset + oplength

        alignopindex = alignopindex + 1

    numquerypositions = len(query_positions)
    refseqlength = len(refseq)

    # use query positions to assess het alleles/covered regions:
    if chromhetsites and ref in chromhetsites:
        desiredhets = chromhetsites[ref]
        for het in desiredhets:
            hetname = het.name
            namefields = hetname.split("_")
            hetpos = int(namefields[-4])
            refallele = namefields[-3]
            altallele = namefields[-2]

            hetstart = het.start
            hetend = het.end
            if hetstart < refstart:
                continue
            if hetend >= refend:
                break
            querystartoffset = query_positions[hetstart-refstart]
            if hetend - refstart + 2 >= len(query_positions):
                queryendoffset = len(queryseq)
            else:
                queryendoffset = query_positions[hetend-refstart+2]
            queryallele = '*'
            if querystartoffset + 1 < queryendoffset - 1:
                queryallele = queryseq[querystartoffset+1:queryendoffset-1]
            alleletype = 'NEITHER'
            if queryallele == refallele:
                alleletype = 'SAME'
            elif queryallele == altallele:
                alleletype = 'ALT'

            if strand == "F":
                querystartcoord = querystart + querystartoffset
                queryendcoord = querystart + queryendoffset - 2
            else:
                querystartcoord = queryend - queryendoffset
                queryendcoord = queryend - querystartoffset

            # adjust end coordinate for cases where there's a deletion in the test assembly at the position of the het:
            if queryendcoord < querystartcoord:
                queryendcoord = querystartcoord

            hetsitealleles[hetname] = {'name':hetname, 'ref':ref, 'refstart':hetstart, 'refend':hetend, 'allele':queryallele, 'query':query, 'start':querystartcoord, 'end':queryendcoord, 'chrom':query}

    return variantlist

def split_aligns_and_sort(splitbamname, bamobj, minindelsize=10000):
    with pysam.AlignmentFile(splitbamname, "wb", header=bamobj.header) as sbfh:
        for align in bamobj.fetch():
            if align.is_secondary:
                continue
            splitaligns = split_align_on_indels(align, minindelsize)
            numaligns = len(splitaligns)
            for splitalign in splitaligns:
                sbfh.write(splitalign)

def split_align_on_indels(align, minindelsize=10000)->list:

    subaligninfo = [] # array of data regarding the sub-alignments

    # general stats about this alignment:
    if align.is_reverse:
        strand = 'R'
    else:
        strand = 'F'
    query = align.query_name
    ref = align.reference_name
    refstart = align.reference_start + 1 # was zero-based, now one-based
    refend = align.reference_end #  was zero-based, adding one for bed format "half open", now one-based
    [left_softclip, right_softclip] = left_right_soft_clip(align)
    [left_hardclip, right_hardclip] = left_right_hard_clip(align)
    originalflag = align.flag
    alignops = align.cigartuples

    logger.debug("Alignment " + align.query_name + " to " + ref + " with soft clip " + str(left_softclip) + "/" + str(right_softclip) + " and hard clip " + str(left_hardclip) + "/" + str(right_hardclip) + " and flag " + str(originalflag))

    # will we hard or soft clip the longest alignment?
    hardcliplongest = False
    if align.is_supplementary or (left_hardclip > 0) or (right_hardclip > 0):
        hardcliplongest = True

    # first position will begin at left-most ref/query base of the alignment (regardless of strand)
    refcurrentoffset = 0
    querycurrentoffset = 0
    reflastalignend = 0 # also zero-based index for the next align segment--this is an offset from refstart
    querylastalignend = 0 # also zero-based index for the next align segment--this is an offset from refstart
    lastalignopindex = 0 # also zero-based index for the next align segment--this is an offset from the first cigar operator
    if left_softclip > 0 or left_hardclip > 0:
        lastalignopindex = 1
    currentalignopindex = 0

    segnumber = 1
    while refcurrentoffset <= refend-refstart and currentalignopindex < len(alignops): # traverse the alignment operator by operator
        alignop = alignops[currentalignopindex]
        op = alignop[0]
        oplength = alignop[1]

        if op in [2, 3]: # deletions
            # if deletion is large enough, append the portion of the alignment to the left as a sub-align, but without the deletion
            if oplength >= minindelsize:
                logger.debug("Align segment " + ref + ":" + str(refstart + reflastalignend) + "-" + str(refstart + refcurrentoffset) + " segment " + str(segnumber))
                logger.debug("Deletion size " + str(op) + " is larger than " + str(minindelsize) + ": splitting")
                segnumber = segnumber + 1
                subalignlength = refcurrentoffset - reflastalignend
                subaligninfo.append({'alignedquerystart':querylastalignend, 'alignedqueryend':querycurrentoffset, 'alignedrefstart':refstart+reflastalignend, 'alignedrefend':refstart+refcurrentoffset, 'cigarops':alignops[lastalignopindex:currentalignopindex], 'subalignlength':subalignlength, 'segnum':segnumber })
                # also need to advance the start coordinate for the next align to point beyond deletion:
                lastalignopindex = currentalignopindex + 1
                reflastalignend = refcurrentoffset + oplength
                querylastalignend = querycurrentoffset
                logger.debug("Setting querylastalignend to " + str(querylastalignend) + " after deletion of length " + str(oplength))

        if op == 1: # insertion
            # if insertion is large enough, append the portion of the alignment to the left as a split align, but without the insertion
            if oplength >= minindelsize:
                logger.debug("Align segment " + ref + ":" + str(refstart + reflastalignend) + "-" + str(refstart + refcurrentoffset) + " to " + str(querylastalignend) + "-" + str(querycurrentoffset) + " segment " + str(segnumber))
                logger.debug("Insertion size " + str(op) + " is larger than " + str(minindelsize) + ": splitting")
                segnumber = segnumber + 1
                subalignlength = refcurrentoffset - reflastalignend
                subaligninfo.append({'alignedquerystart':querylastalignend, 'alignedqueryend':querycurrentoffset, 'alignedrefstart':refstart+reflastalignend, 'alignedrefend':refstart+refcurrentoffset, 'cigarops':alignops[lastalignopindex:currentalignopindex], 'subalignlength':subalignlength, 'segnum':segnumber })
                # also need to advance the start coordinate for the next align to point beyond insertion:
                lastalignopindex = currentalignopindex + 1
                reflastalignend = refcurrentoffset
                querylastalignend = querycurrentoffset + oplength
                logger.debug("Setting querylastalignend to " + str(querylastalignend) + " after insertion of length " + str(oplength))
        currentalignopindex = currentalignopindex + 1

        # advance current positions: cases where reference coord advances (MDN=X):
        if op in [0, 2, 3, 7, 8]:
            refcurrentoffset = refcurrentoffset + oplength
        # cases where query coord advances (MI=X)
        if op in [0, 1, 7, 8]:
            querycurrentoffset = querycurrentoffset + oplength

    logger.debug("Align segment " + ref + ":" + str(refstart + reflastalignend) + "-" + str(refstart + refcurrentoffset) + " to " + str(querylastalignend) + "-" + str(querycurrentoffset) + " last segment " + str(segnumber))
    subalignlength = refcurrentoffset - reflastalignend
    subaligninfo.append({'alignedquerystart':querylastalignend, 'alignedqueryend':querycurrentoffset, 'alignedrefstart':refstart+reflastalignend, 'alignedrefend':refstart+refcurrentoffset, 'cigarops':alignops[lastalignopindex:currentalignopindex], 'subalignlength':subalignlength, 'segnum':segnumber })

    subaligns = create_subalignobjects(align, subaligninfo, hardcliplongest)

    return subaligns

def fix_adjacent_insertions_deletions(cigarops)->list:
    newcigarops = []
    skipnext = False
    # wherever same-size insertions and deletions are adjacent, combine them:
    for i in range(len(cigarops)):
        if skipnext:
            skipnext = False
            continue
        cigartuple = cigarops[i]
        opnum = cigartuple[0]
        oplength = cigartuple[1]
        # if an insertion and the following tuple is a deletion:
        if i < len(cigarops) - 1 and opnum == 1 and cigarops[i+1][0] == 2:
            nextoplength = cigarops[i+1][1]
            skipnext = True
            logger.debug("Combining cigar " + str(opnum) + " length " + str(oplength) + " with cigar " + str(cigarops[i+1][0]) + " length " + str(nextoplength))
            print("Combining cigar " + str(opnum) + " length " + str(oplength) + " with cigar " + str(cigarops[i+1][0]) + " length " + str(nextoplength))
            if nextoplength == oplength:
                newcigarops.append([0, nextoplength])
            elif nextoplength > oplength:
                newcigarops.append([0, oplength])
                newcigarops.append([2, nextoplength - oplength])
            else:
                newcigarops.append([0, nextoplength])
                newcigarops.append([1, oplength - nextoplength])
        # if a deletion and the following tuple is an insertion:
        elif i < len(cigarops) - 1 and opnum == 2 and cigarops[i+1][0] == 1:
            nextoplength = cigarops[i+1][1]
            skipnext = True;
            logger.debug("Combining cigar " + str(opnum) + " length " + str(oplength) + " with cigar " + str(cigarops[i+1][0]) + " length " + str(nextoplength))
            print("Combining cigar " + str(opnum) + " length " + str(oplength) + " with cigar " + str(cigarops[i+1][0]) + " length " + str(nextoplength))
            if nextoplength == oplength:
                newcigarops.append([0, nextoplength])
            elif nextoplength > oplength:
                newcigarops.append([0, oplength])
                newcigarops.append([1, nextoplength - oplength])
            else:
                newcigarops.append([0, nextoplength])
                newcigarops.append([2, oplength - nextoplength])
        # if not adjacent insertion/deletions, simply add this operation
        else:
            newcigarops.append([opnum, oplength])

    return newcigarops

def count_consumed_query(cigarops)->int:

    queryconsumed = 0
    for cigartuple in cigarops:
        opnum = cigartuple[0]
        oplength = cigartuple[1]
        if opnum in [0, 1, 4, 7, 8]:
            queryconsumed = queryconsumed + oplength
    return queryconsumed

def create_subalignobjects(align, subaligninfo:list, hardcliplongest:bool)->list:

    querysamseq = align.query_sequence
    samseqlength = len(querysamseq)
    [full_left_hardclip, full_right_hardclip] = left_right_hard_clip(align)
    subalignobjs = []
    longest = True
    cigarops = align.cigartuples
    for aligninfo in sorted(subaligninfo, key=lambda d: d['subalignlength'], reverse=True):
        newalign = pysam.AlignedSegment()
        newalign.query_name = align.query_name
        newalign.reference_id = align.reference_id
        newalign.mapping_quality = align.mapping_quality
        newalign.reference_start = aligninfo['alignedrefstart'] - 1
        cigartuples = fix_adjacent_insertions_deletions(aligninfo['cigarops'])
        oldqueryconsumed = count_consumed_query(aligninfo['cigarops'])
        newqueryconsumed = count_consumed_query(cigartuples)
        if oldqueryconsumed != newqueryconsumed:
            logger.info("Cigar op length change for " + align.query_name)
            logger.info("Old length: " + str(oldqueryconsumed) + "\nNew length: " + str(newqueryconsumed))

        # soft clip the longest unless hardcliplongest was specified
        if longest and not hardcliplongest:
            logger.debug("Will softclip align of length " + str(aligninfo['subalignlength']) + " with start " + str(aligninfo['alignedquerystart']))
            newalign.query_sequence = querysamseq
            leftsoftclip = align.query_alignment_start + aligninfo['alignedquerystart']
            rightsoftclip = samseqlength - aligninfo['alignedqueryend'] - align.query_alignment_start
            logger.debug("Will add " + str(leftsoftclip) + " left softclip and right softclip " + str(rightsoftclip) + " to align of length " +str(aligninfo['subalignlength']) + " with sam length " + str(samseqlength))
            if leftsoftclip > 0:
                cigarlist = [(4, leftsoftclip)] + cigartuples
            else:
                cigarlist = cigartuples
            if rightsoftclip > 0:
                cigarlist.append([4, rightsoftclip])
            try:
                newalign.cigartuples = cigarlist
            except OverflowError:
                logger.critical("Overflow Error while retrieving supplementary CIGAR string. Read name: {0}, Position: {1}, CIGAR: {2}".format(align.query_name, str(aligninfo['alignedquerystart']), str(aligninfo['alignedrefstart'])))
                continue
            #for cigar in cigarlist:
                #logger.debug("CIGAR\t" + str(cigar[0]) + ":" + str(cigar[1]))
            queryconsumedbases = count_consumed_query(cigarlist)
            logger.debug("Query seq length is " + str(samseqlength) + " and cigar consumes " + str(queryconsumedbases))
        else:
            logger.debug("Will hardclip align of length " + str(aligninfo['subalignlength']))
            lefthardclip = align.query_alignment_start + aligninfo['alignedquerystart'] + full_left_hardclip
            righthardclip = samseqlength - aligninfo['alignedqueryend'] - align.query_alignment_start + full_right_hardclip
            logger.debug("Will add " + str(lefthardclip) + " left hardclip and right hardclip " + str(righthardclip) + " to align of length " +str(aligninfo['subalignlength']) + " with sam length " + str(samseqlength))
            if lefthardclip > 0:
                cigarlist = [(5, lefthardclip)] + cigartuples
            else:
                cigarlist = cigartuples
            if righthardclip > 0:
                cigarlist.append([5, righthardclip])
            try:
                newalign.cigartuples = cigarlist
            except OverflowError:
                logger.critical("Overflow Error while retrieving supplementary CIGAR string. Read name: {0}, Position: {1}, CIGAR: {2}".format(align.query_name, str(aligninfo['alignedquerystart']), str(aligninfo['alignedrefstart'])))
                continue
            query_seqstart = aligninfo['alignedquerystart'] + align.query_alignment_start
            query_seqend = aligninfo['alignedqueryend'] + align.query_alignment_start
            newalign.query_sequence = querysamseq[query_seqstart:query_seqend]
            queryseqlength = query_seqend - query_seqstart
            #logger.debug("Query sequence has length " + str(lefthardclip) + " hardclip and right hardclip " + str(righthardclip) + " to align of length " +str(aligninfo['subalignlength']))
            queryconsumedbases = count_consumed_query(cigarlist)
            logger.debug("Query seq length is " + str(queryseqlength) + " and cigar consumes " + str(queryconsumedbases))

        if not longest:
            newalign.flag = align.flag | 2048
            logger.debug("Marking align of length " + str(aligninfo['subalignlength']) + " as supplementary")
        else:
            newalign.flag = align.flag
        longest = False
        subalignobjs.append(newalign)

    return subalignobjs

def exclude_variants(variants:list, excludedregionsobj:pybedtools.BedTool)->list:
    # create bedintervals for variants:
    logger.debug("Excluding lots of variants in excluded regions")
    numvariants = len(variants)
    logger.debug("There are " + str(numvariants) + " variants")
    variantbedstringlist = []
    for variant in variants:
        chrom = variant.chrom
        start = variant.start
        end = variant.end
        name = variant.name
        variantbedstringlist.append(chrom + "\t" + str(start) + "\t" + str(end) + "\t" + name + "\n")

    variantbedstring = ''.join(variantbedstringlist)
    variantbedobj = pybedtools.BedTool(variantbedstring, from_string=True)
    logger.debug("Created BedTool object")
    excludedvariants = bedtoolslib.intersectintervals(variantbedobj, excludedregionsobj, wa=True)
    excludeddict = {}
    for excludedvariant in excludedvariants:
        excludeddict[excludedvariant.name] = True

    newvariants = []
    numexcluded = 0
    numretained = 0
    for variant in variants:
        if variant.name in excludeddict.keys():
            newvariants.append(varianttuple(chrom=variant.chrom, start=variant.start, end=variant.end, name=variant.name, vartype=variant.vartype, excluded = True, qvscore=variant.qvscore))
            numexcluded = numexcluded + 1
        else:
            newvariants.append(variant)
            numretained = numretained + 1

    logger.debug("Excluded " + str(numexcluded) + " variants and kept " + str(numretained) + " variants")

    return newvariants

def read_paf_aligns(paffile:str, mintargetlength=0)->list:

    alignlist = []
    
    paf = Path(paffile)
    if not paf.is_file():
        logger.critical("PAF file " + paffile + " must exist and be readable")
        exit(1)

    with open(paffile, "r") as pfh:
        alignline = pfh.readline()
        while alignline:
            alignline = alignline.rstrip()
            fields = alignline.split("\t")
            if len(fields) >= 12:
                [query, querylength, querystartzb, queryend, strand, target, targetlength, targetstartzb, targetend, resmatches, blocklength, mapqual] = fields[0:12]
            else:
                logger.critical("Input paf-file has fewer than 12 tab-delimited columns. Unable to process.")
                exit(1)

            #if len(fields) > 12 and len(list(filter(lambda x:'tp:A:S' in x, fields[12:]))) > 0:
                #continue

            querystart = int(querystartzb) + 1
            queryend = int(queryend)
            queryalignlength = queryend - int(querystartzb)
            targetstart = int(targetstartzb) + 1
            targetend = int(targetend)
            targetalignlength = targetend - int(targetstartzb)

            if targetalignlength >= mintargetlength:
                if strand == "+":
                    alignlist.append({'query':query, 'querylength':int(querylength), 'querystart':querystart, 'queryend':queryend, 'queryalignlength':queryalignlength, 'strand':strand, 'target':target, 'targetlength':int(targetlength), 'targetstart':targetstart, 'targetend':targetend, 'targetalignlength':targetalignlength})
                else:
                    alignlist.append({'query':query, 'querylength':int(querylength), 'queryend':querystart, 'querystart':queryend, 'queryalignlength':queryalignlength, 'strand':strand, 'target':target, 'targetlength':int(targetlength), 'targetstart':targetstart, 'targetend':targetend, 'targetalignlength':targetalignlength})
            alignline = pfh.readline()

    return alignlist

def read_bam_aligns(bamobj, mintargetlength=0)->list:

    alignlist = []

    refentries = bamobj.references
    reflengths = bamobj.lengths
    reflengthdict = {}
    for tid in range(len(refentries)):
        reflengthdict[refentries[tid]] = reflengths[tid]

    for align in bamobj.fetch():
        if align.is_secondary:
            continue
        if align.reference_length >= mintargetlength:
            query, querystart, queryend, ref, refstart, refend, strand = retrieve_align_data(align)
            querylength = align.query_length
            targetlength = reflengthdict[ref]
            if strand == "F":
                queryleft = querystart
                queryright = queryend
                alignlist.append({'query':query, 'querylength':querylength, 'querystart':querystart, 'queryend':queryend, 'queryalignlength':queryend-querystart+1, 'strand':'+', 'target':ref, 'targetlength':targetlength, 'targetstart':refstart, 'targetend':refend, 'targetalignlength':refend-refstart+1})
            else:
                queryleft = queryend
                queryright = querystart
                alignlist.append({'query':query, 'querylength':querylength, 'queryend':querystart, 'querystart':queryend, 'queryalignlength':queryend-querystart+1, 'strand':'-', 'target':ref, 'targetlength':targetlength, 'targetstart':refstart, 'targetend':refend, 'targetalignlength':refend-refstart+1})

    return alignlist

# this routine assumes query start > query end for reverse strand alignments
def assess_overall_structure(aligndata:list, refobj, queryobj, outputfiles, bedobjects, benchmark_stats, args):

    # create a "clustercoverage" dictionary that will contain each chromosome's number of clusters and bases covered
    benchmark_stats["clustercoverage"] = {}
    benchmark_stats["alignclusters"] = {}
    benchmark_stats["numnonexcludedbases"] = {}

    # maximum separating distance along target to include in one cluster of alignments
    maxdistance = args.maxclusterdistance

    # create a directory for alignment line plots for each chromosome:
    output.create_output_directory(outputfiles["alignplotdir"])
    alignplotprefix = outputfiles["alignplotprefix"]

    # calculate number of non-excluded bases for each ref entry:
    # since allexcludedregions bed object is merged,can subtract out excluded bases interval by interval
    for ref in refobj.references:
        benchmark_stats["numnonexcludedbases"][ref] = refobj.get_reference_length(ref)
    if bedobjects["allexcludedregions"] is not None:
        for interval in bedobjects["allexcludedregions"]:
            ref = interval.chrom
            if ref in benchmark_stats["numnonexcludedbases"].keys():
                benchmark_stats["numnonexcludedbases"][ref] = benchmark_stats["numnonexcludedbases"][ref] - len(interval)
    
    aligndict = {}
    for align in aligndata:
        refentry = align["target"]
        if refentry not in aligndict:
            aligndict[refentry] = [align]
        else:
            aligndict[refentry].append(align)

    # assess each benchmark entry, one by one
    for refentry in sorted(aligndict.keys()):
        logger.debug("Entry " + refentry + " start")
        refnelength = benchmark_stats["numnonexcludedbases"][refentry]
        refalignclusters = []
        # sort alignments from longest (along the benchmark) to shortest:
        numaligns = len(aligndict[refentry])
        logger.debug("Sorting " + str(numaligns) + " " + refentry + " aligns")
        aligndict[refentry].sort(reverse=True, key=lambda align: align["targetalignlength"])
        # calculate slope as query diff over ref diff, and cluster alignments within the same query/target band:
        logger.debug("Adding " + str(numaligns) + " " + refentry + " aligns to clusters")
        for refalign in aligndict[refentry]:
            alignrefstart = refalign['targetstart']
            alignrefend = refalign['targetend']
            alignquery = refalign['query']
            alignquerystart = refalign['querystart']
            alignqueryend = refalign['queryend']
            add_align_to_clusters(refalign, refalignclusters, maxdistance)

        # split clusters that are separated along the target by more than maxdistance:
        logger.debug("Splitting " + refentry + " clusters")
        refalignclusters = split_disjoint_clusters(refalignclusters, maxdistance)

        logger.debug("Forming " + refentry + " cluster bed string")
        for cluster in sorted(refalignclusters, key=lambda c: c["aligns"][0]["targetstart"]):
            clusterquery = cluster["query"] 
            clusterslope = cluster["slope"]
            clusterintercept = cluster["intercept"]
            clusterbedstring = ""
            for align in sorted(cluster["aligns"], key=lambda a: (a["targetstart"], a["targetend"])):
                clusterbedstring = clusterbedstring + refentry + "\t" + str(align["targetstart"]) + "\t" + str(align["targetend"]) + "\t" + clusterquery + "_" + str(align["querystart"]) + "_" + str(align["queryend"]) + "\n"

            bedtool = pybedtools.BedTool(clusterbedstring, from_string = True)
            mergedbedtool = bedtoolslib.mergeintervals(bedtool)
            clusterbases = bedtoolslib.bedsum(mergedbedtool)
            if bedobjects["allexcludedregions"] is not None:
                nonexcludedbedtool = bedtoolslib.subtractintervals(mergedbedtool, bedobjects["allexcludedregions"])
            else:
                nonexcludedbedtool = mergedbedtool
            cluster["nonexcludedcoveredbases"] = bedtoolslib.bedsum(nonexcludedbedtool)
            logger.debug("Cluster on " + clusterquery + " has " + str(clusterbases) + " non-redundant bases, " + str(cluster["nonexcludedcoveredbases"]) + " of which are not excluded")

        logger.debug("See how many " + refentry + " clusters are needed to cover 95% of ref")
        # calculate how many clusters needed to cover 95% of ref:
        totalnonexcludedcovered = 0
        clustercount = 0
        lca95 = None
        nca95 = None
        clusterno = 1
        refentrybedstring = ""
        for cluster in sorted(refalignclusters, key=lambda c:c["nonexcludedcoveredbases"], reverse = True):
            clusterquery = cluster["query"] 
            for align in sorted(cluster["aligns"], key=lambda a: (a["targetstart"], a["targetend"])):
                if lca95 is None:
                    clustername = "Cluster" + str(clusterno)
                else:
                    clustername = "SmallCluster" + str(clusterno)
                refentrybedstring = refentrybedstring + refentry + "\t" + str(align["targetstart"]) + "\t" + str(align["targetend"]) + "\t" + clusterquery + "_" + str(align["querystart"]) + "_" + str(align["queryend"]) + "_" + clustername + "\n"
            clusterno = clusterno + 1

            nebases = cluster["nonexcludedcoveredbases"]
            logger.debug("Cluster " + clustername + " has " + str(nebases) + " nonexcluded covered bases")

            if lca95 is None:
                totalnonexcludedcovered = totalnonexcludedcovered + cluster["nonexcludedcoveredbases"]
            clustercount = clustercount + 1
            if lca95 is None and totalnonexcludedcovered > 0.95*refnelength:
                lca95 = clustercount
                nca95 = cluster["nonexcludedcoveredbases"]

        logger.debug("Saving BED file for " + refentry + " clusters")
        refentrybedtool = pybedtools.BedTool(refentrybedstring, from_string = True)
        sortedrefentrybedtool = refentrybedtool.sort()
        sortedrefentrybedtool.saveas(alignplotprefix + "." + refentry + ".clusters.bed")

        # Note: lca95 will be "None" for reference entries not able to be covered
        benchmark_stats["clustercoverage"][refentry] = {"lca95":lca95, "nonexcludedcovered":totalnonexcludedcovered}
        refalignclusters.sort(key=lambda c: c["aligns"][0]["targetstart"])
        benchmark_stats["alignclusters"][refentry] = refalignclusters
    
    return benchmark_stats

def left_right_soft_clip(align)->list:
    leftsoftclip = 0
    rightsoftclip = 0
    
    if align.cigartuples[0][0]==4:
        leftsoftclip = align.cigartuples[0][1]
    if align.cigartuples[-1][0]==4:
        rightsoftclip = align.cigartuples[-1][1]

    return [leftsoftclip, rightsoftclip]

def left_right_hard_clip(align)->list:
    lefthardclip = 0
    righthardclip = 0
    
    if align.cigartuples[0][0]==5:
        lefthardclip = align.cigartuples[0][1]
    if align.cigartuples[-1][0]==5:
        righthardclip = align.cigartuples[-1][1]

    return [lefthardclip, righthardclip]

# Routine to retrieve alignment ref coords and cigar ops for subalignment limits (specified by offsets from query_alignment_start)
def retrieve_refcoords_and_cigars_from_querycoords(align, desiredquerystart:int, desiredqueryend:int):

    # general stats about this alignment:
    query, querystart, queryend, ref, refstart, refend, strand = retrieve_align_data(align)
    #print("Align " + query + ":" + str(querystart) + "-" + str(queryend) + " looking for positions " + str(desiredquerystart) + " and " + str(desiredqueryend))
    [left_softclip, right_softclip] = left_right_soft_clip(align)
    [left_hardclip, right_hardclip] = left_right_hard_clip(align)
    alignops = align.cigartuples

    # first position will begin at left-most ref/query base of the alignment (regardless of strand)
    refcurrentoffset = 0
    querycurrentoffset = 0
    currentalignopindex = 0

    desiredrefstart = None
    desiredrefend = None
    desiredcigarops = []
    desiredfirstopindex = None
    desiredlastopindex = None
    while refcurrentoffset <= refend-refstart and currentalignopindex < len(alignops): # traverse the alignment operator by operator
        alignop = alignops[currentalignopindex]
        op = alignop[0]
        oplength = alignop[1]

        currentalignopindex = currentalignopindex + 1

        if op in [0, 7, 8]: # substitution
            matches_to_add = oplength
            if desiredquerystart >= querycurrentoffset and desiredquerystart < querycurrentoffset + oplength:
                offsetwithinblock = desiredquerystart - querycurrentoffset
                desiredrefstart = refcurrentoffset + refstart + offsetwithinblock
                matches_to_add = matches_to_add - offsetwithinblock 
            if desiredqueryend >= querycurrentoffset and desiredqueryend < querycurrentoffset + oplength:
                offsetwithinblock = desiredqueryend - querycurrentoffset
                desiredrefend = refcurrentoffset + refstart + offsetwithinblock
                matches_to_add = matches_to_add - oplength + offsetwithinblock 
            if desiredrefstart is not None:
                desiredcigarops.append([op, matches_to_add])
            if desiredrefend is not None:
                return desiredquerystart, desiredqueryend, ref, desiredrefstart, desiredrefend, desiredcigarops
            refcurrentoffset = refcurrentoffset + oplength
            querycurrentoffset = querycurrentoffset + oplength
        # advance current positions: cases where reference coord advances (DN):
        elif op in [2, 3]:
            if desiredrefstart is not None:
                desiredcigarops.append([op, oplength])
            refcurrentoffset = refcurrentoffset + oplength
        # cases where query coord advances (I)
        # BUG: need to account here for cases where desired query endpoints are reached within an insertion
        elif op == 1:
            insertions_to_add = oplength
            if desiredquerystart >= querycurrentoffset and desiredquerystart < querycurrentoffset + oplength:
                desiredrefstart = refcurrentoffset + refstart
                insertions_to_add = insertions_to_add + querycurrentoffset - desiredquerystart
            if desiredqueryend >= querycurrentoffset and desiredqueryend < querycurrentoffset + oplength:
                desiredrefend = refcurrentoffset + refstart
                insertions_to_add = desiredqueryend - querycurrentoffset

                desiredcigarops.append([op, insertions_to_add])

                return desiredquerystart, desiredqueryend, ref, desiredrefstart, desiredrefend, desiredcigarops

            if desiredrefstart is not None:
                desiredcigarops.append([op, insertions_to_add])
            querycurrentoffset = querycurrentoffset + oplength

        if desiredrefstart is not None and desiredrefend is not None:
            return desiredquerystart, desiredqueryend, ref, desiredrefstart, desiredrefend, desiredcigarops

    return desiredquerystart, desiredqueryend, ref, desiredrefstart, desiredrefend, desiredcigarops

# do a quick comparison of two alignments to see if one is contained within the other, or if they are dovetail to the left, right, or both
# function returns either: ref start/end within the first alignment where the second alignment has agreeing endpoints, but -1 for 
# start or end if there is no agreement at that end
def compare_alignments(align1, align2, switched=False):

    # coordinates are all one-based, with start at beginning of *original* sequence (not left end of the alignment)
    query1, query1start, query1end, ref1, ref1start, ref1end, strand1 = retrieve_align_data(align1)
    query2, query2start, query2end, ref2, ref2start, ref2end, strand2 = retrieve_align_data(align2)

    logger.debug("Comparing " + ref1 + ":" + str(ref1start) + "-" + str(ref1end) + " to " + query2 + ":" + str(query2start) + "-" + str(query2end) + "/" + strand1)

    if strand1 != strand2:
        return [-1, -1]

    # for soft clipped alignments, query positions include soft clipped bases (so are larger by the amount of the soft clipping)
    align1pairs = align1.get_aligned_pairs(matches_only=True)
    align2pairs = align2.get_aligned_pairs(matches_only=True)

    [leftsoftclip1, rightsoftclip1] = left_right_soft_clip(align1)
    [leftsoftclip2, rightsoftclip2] = left_right_soft_clip(align2)

    # for align1, ref positions in array are the second tuple element + ref1start, query pos are first tuple element + query1start
    align1index = 0
    align2index = 0
    alignoverlaprefstart = None
    alignoverlapquerystart = None
    lastoverlaprefstart = None
    lastoverlapquerystart = None
    while align1index < len(align1pairs) and align2index < len(align2pairs):
        ref1pos = align1pairs[align1index][1] + 1
        if strand1 == 'F':
            query1pos = align1pairs[align1index][0] + query1start
            if leftsoftclip1 > 0:
                query1pos = query1pos - leftsoftclip1
        else:
            query1pos = align1pairs[align1index][0] + query1start
            if rightsoftclip1 > 0:
                query1pos = query1pos - rightsoftclip1
    
        if not switched:
            ref2pos = align2pairs[align2index][1] + 1
            if strand2 == 'F':
                query2pos = align2pairs[align2index][0] + query2start
                if leftsoftclip2 > 0:
                    query2pos = query2pos - leftsoftclip2
            else:
                query2pos = query2start + align2pairs[align2index][0]
                if rightsoftclip2 > 0:
                    query2pos = query2pos - rightsoftclip2

        else: # need to switch ref/query
            if strand2 == 'F':
                ref2pos = align2pairs[align2index][0] + query2start
                query2pos = align2pairs[align2index][1] + 1
                if leftsoftclip2 > 0:
                    ref2pos = ref2pos - leftsoftclip2
            else:
                ref2pos = align2pairs[align2index][0] + query2start
                #query2pos = ref2end - align2pairs[align2index][1]
                query2pos = align2pairs[align2index][1] + 1
                if rightsoftclip2 > 0:
                    ref2pos = ref2pos - rightsoftclip2

        # based on ref positions, should we advance one index, the other, or both?
        #if ref1start==21156327 and ref1end==34675573:
            #print(str(ref1pos) + "\t" + str(ref2pos) + "\t" + str(query1pos) + "\t" + str(query2pos) + "\t" + str(align1index) + "\t" + str(align2index))
        if ref1pos < ref2pos:
            align1index = align1index + 1
        elif ref1pos > ref2pos:
            align2index = align2index + 1
        elif ref1pos == ref2pos:
            if query1pos == query2pos:
                if alignoverlaprefstart is None:
                    alignoverlaprefstart = ref1pos
                    alignoverlapquerystart = query1pos
                lastoverlaprefstart = ref1pos
                lastoverlapquerystart = query1pos
            align1index = align1index + 1
            align2index = align2index + 1

    return [alignoverlaprefstart, lastoverlaprefstart]

def add_align_to_clusters(align:dict, alignclusters:list, maxdistance:int):

    alignstart = align['targetstart']
    alignend = align['targetend']
    alignquery = align['query']
    alignquerystart = align['querystart']
    alignqueryend = align['queryend']
    alignslope = (alignend - alignstart)/(alignqueryend-alignquerystart)
    alignintercept = alignstart - int(alignslope * alignquerystart)

    # try to assign this align to a pre-existing cluster of aligns:
    assigned = False
    for cluster in alignclusters:
        if cluster["query"] != alignquery:
            continue
        clusterquery = cluster["query"]
        clusterslope = cluster["slope"]
        clusterintercept = cluster["intercept"]
        predstart = clusterintercept + clusterslope * alignquerystart
        if abs(predstart - alignstart) <= maxdistance:
            cluster["aligns"].append(align)
            assigned = True
            break

    # create a new cluster if none were appropriate
    if not assigned:
        alignclusters.append({'query':alignquery, 'slope':alignslope, 'intercept':alignintercept, 'aligns':[align]})

    return 0

def split_disjoint_clusters(refalignclusters:list, maxdistance:int)->list:

    spinoffclusters = []
    for cluster in refalignclusters:
        if len(cluster["aligns"])==1:
            continue

        runningalignlist = []
        maxposition = 0
        for align in sorted(cluster["aligns"], key=lambda a: (a["targetstart"], a["targetend"])):
            # do we have too big a gap?
            if len(runningalignlist) > 0 and align["targetstart"] - maxposition > maxdistance:
                spinoffclusters.append({'query':cluster["query"], 'slope':cluster["slope"], 'intercept':cluster["intercept"], 'aligns':runningalignlist})
                runningalignlist = []
                maxposition = 0
                logger.debug("Splitting align cluster that ends at " + cluster["query"] + ":" + str(maxposition) + " for new align starting at " + str(align["targetstart"]))
            runningalignlist.append(align)
            maxposition = max(maxposition, align["targetend"])
        cluster["aligns"] = runningalignlist

    return refalignclusters + spinoffclusters

def trim_bamfile_to_intervals(bamfile, intervals, outputbam, headerbam, args, sort=True, index=True):

    [aligns, alignedintervals] = index_aligns_by_boundaries(bamfile, args)
    subaligns = find_phaseblock_subaligns(intervals, alignedintervals, aligns)
    write_aligns_to_bamfile(outputbam, subaligns, headerbam=headerbam, sort=sort, index=index)

def index_aligns_by_boundaries(bamfile, args):

    logger.info("Indexing aligns in " + bamfile)
    alignobj = pysam.AlignmentFile(bamfile, "rb")
    aligndict = {}
    alignbedstring = ""
    for align in alignobj.fetch():
        if align.is_secondary:
            continue
        if align.reference_length >= args.minalignlength:
            query, querystart, queryend, ref, refstart, refend, strand = retrieve_align_data(align)
            alignname = query + "_" + str(querystart) + "_" + str(queryend)
            alignbedstring = alignbedstring + query + "\t" + str(querystart) + "\t" + str(queryend) + "\t" + str(alignname) + "\n"
            aligndict[alignname] = align

    alignbedtool = pybedtools.BedTool(alignbedstring, from_string = True)

    return [aligndict, alignbedtool]

# Trim alignments to the boundaries of phase blocks:
def find_phaseblock_subaligns(phaseblockints, alignedintervals, aligndict):

    # For each alignment, create a list of intersected intervals representing phaseblock regions within the alignment
    # Intervals are in the direction of the query, even if the alignment is on the reverse strand
    phasedaligndict = {}
    phaseblockalignints = bedtoolslib.intersectintervals(phaseblockints, alignedintervals, wo=True)
    numphaseblockaligns = len(phaseblockalignints)
    logger.debug("There are " + str(numphaseblockaligns) + " aligned intervals within phase block intervals")
    for blockalign in phaseblockalignints:
        alignname = blockalign[12]
        if alignname not in phasedaligndict.keys():
            phasedaligndict[alignname] = []
        phasedaligndict[alignname].append(blockalign)

    # Now for each aligned region, sort the intersecting intervals from low to high coordinate:
    sortedaligns = sorted(phasedaligndict.keys())
    subalignlist = []
    for alignname in sortedaligns:
        if alignname not in aligndict.keys():
            logger.debug("Skipping alignment " + alignname + "--not in align dictionary!")
            print("Skipping alignment " + alignname + "--not in align dictionary!")
            continue
        alignobj = aligndict[alignname]
        query, querystart, queryend, ref, refstart, refend, strand = retrieve_align_data(alignobj)
        phaseblockintersects = phasedaligndict[alignname]
        subaligninfo = []
        segnumber = 1
        logger.debug("Processing align " + alignname)
        for phaseblock in sorted(phaseblockintersects, key=lambda x:(x.start, x.stop)):
            pbstart = phaseblock.start
            pbend = phaseblock.end
            alignstart = int(phaseblock[10])
            alignend = int(phaseblock[11])
            intersectlength = int(phaseblock[13])
            intersectstart = max(pbstart, alignstart)
            intersectend = min(pbend, alignend)
            logger.debug("Align " + alignname + " intersects phase block " + str(pbstart) + "-" + str(pbend) + " from " + str(intersectstart) + " to " + str(intersectend))
            
            if (intersectend - intersectstart) != intersectlength:
                print(str(intersectlength) + " is not equal to " + str(intersectend) + "-" + str(intersectstart))
            # Need to calculate the correct aligned query start/end for the routine "create_subalignobjects":
            if strand == 'F':
                querystartoffset = intersectstart - querystart
                queryendoffset = intersectend - querystart
            else: # this should be corrected:
                querystartoffset = queryend - intersectend
                queryendoffset = queryend - intersectstart

            desiredquerystart, desiredqueryend, desiredref, desiredrefstart, desiredrefend, desiredcigarops = retrieve_refcoords_and_cigars_from_querycoords(alignobj, querystartoffset, queryendoffset)

            desiredcigarqueryconsumed = count_consumed_query(desiredcigarops)
            if desiredcigarqueryconsumed != intersectlength:
               print("Align " + alignname + " intersects phase block " + str(pbstart) + "-" + str(pbend) + " from " + str(intersectstart) + " to " + str(intersectend))
               print("Found refcoords " + desiredref + ":" + str(desiredrefstart) + "-" + str(desiredrefend))
               print("Cigar ops consume " + str(desiredcigarqueryconsumed) + " bases and intersect is " + str(intersectlength))
               continue

            if desiredrefstart is None or desiredrefend is None:
                print("Found refcoords " + desiredref + ":" + str(desiredrefstart) + "-" + str(desiredrefend))
                print("Unable to find reference start/end for query " + str(querystartoffset) + "-" + str(queryendoffset) + " in align " + alignname)
                continue
            subaligninfo.append({'alignedquerystart':querystartoffset, 'alignedqueryend':queryendoffset, 'alignedrefstart':desiredrefstart, 'alignedrefend':desiredrefend, 'cigarops':desiredcigarops, 'subalignlength':intersectlength, 'segnum':segnumber})
            segnumber = segnumber + 1

        [left_hardclip, right_hardclip] = left_right_hard_clip(alignobj)
        hardcliplongest = False
        if alignobj.is_supplementary or (left_hardclip > 0) or (right_hardclip > 0):
            hardcliplongest = True
        subalignobjs = create_subalignobjects(alignobj, subaligninfo, hardcliplongest)
        subalignlist = subalignlist + subalignobjs

    return subalignlist

def write_aligns_to_bamfile(bamfilename:str, aligns:list, headerbam:str="test_bam_for_header.bam", sort=True, index=True):

    bamobj = pysam.AlignmentFile(headerbam, "r")

    with pysam.AlignmentFile(bamfilename, "wb", header=bamobj.header) as sbfh:
        for align in aligns:
            sbfh.write(align)

    if sort:
        sortedbamfilename = bamfilename.replace(".bam", ".sort.bam")
        pysam.sort("-o", sortedbamfilename, bamfilename)
        if index:
            pysam.index(sortedbamfilename)
        return sortedbamfilename
    else:
        return bamfilename

def write_aligns_to_samfile(samfilename:str, aligns:list, headerbam:str="test_bam_for_header.bam"):

    bamobj = pysam.AlignmentFile(headerbam, "r")

    print("Writing alignments to " + samfilename)
    with pysam.AlignmentFile(samfilename, "w", header=bamobj.header) as sbfh:
        for align in aligns:
            sbfh.write(align)


def merge_trimmed_bamfiles(mattrimmedbamfile:str, pattrimmedbamfile:str, benchdiploidheaderstring:str, outputfiles:dict, sort=True):

    mergedtrimmedsamfile = outputfiles["trimmedphasedalignprefix"] + ".merge.dipheader.sam"
    with open(mergedtrimmedsamfile, "w") as sfh:
        sfh.write(benchdiploidheaderstring)

    logger.debug("Printed diploid header string to " + mergedtrimmedsamfile)

    mattrimmedsamreturnval = os.system("samtools view " + mattrimmedbamfile + " >> " + mergedtrimmedsamfile)
    logger.debug("Writing maternal trimmed sam alignments to " + mergedtrimmedsamfile + " had return value " + str(mattrimmedsamreturnval))
    pattrimmedsamreturnval = os.system("samtools view " + pattrimmedbamfile + " >> " + mergedtrimmedsamfile)
    logger.debug("Writing paternal trimmed sam alignments to " + mergedtrimmedsamfile + " had return value " + str(pattrimmedsamreturnval))
    mergedtrimmedbamfile = outputfiles["trimmedphasedalignprefix"] + ".merge.bam"
    samtobamreturnval = os.system("samtools view -bS " + mergedtrimmedsamfile + " -o " + mergedtrimmedbamfile)
    logger.debug("Converting merged trimmed sam alignments to bam had return value " + str(samtobamreturnval))
    if pattrimmedsamreturnval == 0 and mattrimmedsamreturnval == 0 and samtobamreturnval == 0:
        logger.debug("Removing mat/pat trimmed bam files and sam alignments")
        os.remove(mergedtrimmedsamfile)
        os.remove(mattrimmedbamfile)
        os.remove(pattrimmedbamfile)
    else:
        logger.critical("Error merging mat/pat trimmed bam files to a single bam file--exiting")
        exit(1)

    if sort:
        trimmedphasedsortbam = outputfiles["trimmedphasedalignprefix"] + ".merge.sort.bam"
        trimmedbamsortreturnval = os.system("samtools sort " + mergedtrimmedbamfile + " > " + trimmedphasedsortbam)
        logger.debug("Sorting merged trimmed bam alignments had return value " + str(trimmedbamsortreturnval))
        trimmedbamindexval = os.system("samtools index " + trimmedphasedsortbam)
        logger.debug("Indexing trimmed bam file had return value " + str(trimmedbamindexval))
        if trimmedbamsortreturnval==0 and trimmedbamindexval==0:
            os.remove(mergedtrimmedbamfile)
        else:
            logger.critical("Error sorting trimmed merged bam file " + mergedtrimmedbamfile)
            exit(1)
        return trimmedphasedsortbam

    return mergedtrimmedbamfile




import pytest
import subprocess
import sys
import os
import pysam
from GQC import bench
from GQC import output
from GQC import seqparse
from GQC import alignparse
from GQC import mummermethods
from GQC import bedtoolslib

def test_checkforprogs():
    bench.check_for_bedtools()
    no_rscript = bench.check_for_R()

#def test_createoutputdir():
    #args = bench.parse_arguments(['-c', 'GQC/benchconfig.txt', '-b', 'tests/test.sort.bam', '-r', 'tests/testbenchmark.fasta.gz', '-q', 'tests/testassembly.fasta.gz', '-p', 'tests/testrun'])
    #configvals = bench.read_config_data(args)
    #outputdir = output.create_output_directory(args.prefix)
    #assert os.path.isdir(outputdir)

def test_writebedfiles():
    args = bench.parse_arguments(['-c', 'GQC/benchconfig.txt', '-b', 'tests/test.sort.bam', '-r', 'tests/testbenchmark.fasta.gz', '-q', 'tests/testassembly.fasta.gz', '-p', 'tests/testrun'])
    configvals = bench.read_config_data(args)

    refobj = pysam.FastaFile(args.reffasta)
    queryobj = pysam.FastaFile(args.queryfasta)
    outputfiles = output.name_output_files(args, args.prefix)
    seqparse.write_genome_bedfiles(queryobj, refobj, args, configvals, outputfiles, {})

    with open('tests/testrun/genome.test.bed', 'r') as gfh:
        fields = gfh.readline().rstrip().split()
        assert(int(fields[2]) == 9721)
    with open('tests/testrun/atgcseq.test.bed', 'r') as gfh:
        fields = gfh.readline().rstrip().split()
        assert(int(fields[2]) == 9721)

    alignobj = pysam.AlignmentFile(args.bam, "rb")
    aligndata = alignparse.read_bam_aligns(alignobj, args.minalignlength)
    rlis_aligndata = mummermethods.filter_aligns(aligndata, "target")

    assert(len(aligndata) == 1)

